from Acquisition import aq_base
from Acquisition import aq_chain
from plone.app.multilingual.content.lrf import ILanguageRootFolder
from plone.app.multilingual.interfaces import ILanguageIndependentFolder
from plone.uuid.interfaces import ATTRIBUTE_NAME
from plone.uuid.interfaces import IAttributeUUID
from plone.uuid.interfaces import IUUID
from zope.component import adapter
from zope.interface import implementer


@implementer(IUUID)
@adapter(ILanguageRootFolder)
def lrfUUID(context):
    return getattr(aq_base(context), ATTRIBUTE_NAME, None)


@implementer(IUUID)
@adapter(ILanguageIndependentFolder)
def lifUUID(context):
    return getattr(aq_base(context), ATTRIBUTE_NAME, None)


@implementer(IUUID)
@adapter(IAttributeUUID)
def attributeUUID(context):
    is_language_independent = False
    for element in aq_chain(context):
        if ILanguageIndependentFolder.providedBy(element):
            is_language_independent = True
        if ILanguageRootFolder.providedBy(element) and is_language_independent:
            uid = getattr(aq_base(context), ATTRIBUTE_NAME, None) or ""
            return uid + "-" + element.id if uid is not None else None
    return getattr(context, ATTRIBUTE_NAME, None)



from Acquisition import aq_inner
from plone.app.i18n.locales.browser.selector import LanguageSelector
from plone.app.multilingual.browser.selector import LanguageSelectorViewlet
from plone.app.multilingual.dx.interfaces import ILanguageIndependentField
from plone.app.multilingual.interfaces import IMultiLanguageExtraOptionsSchema
from plone.dexterity.browser.edit import DefaultEditForm
from plone.registry.interfaces import IRegistry
from plone.z3cform import layout
from Products.CMFCore.utils import getToolByName
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from zope.component import getUtility


class MultilingualEditForm(DefaultEditForm):
    babel = ViewPageTemplateFile("templates/dexterity_edit.pt")

    def languages(self):
        """Deprecated"""
        context = aq_inner(self.context)

        ls = LanguageSelector(context, self.request, None, None)
        ls.update()
        results = ls.languages()

        supported_langs = [v["code"] for v in results]
        missing = {str(c) for c in supported_langs}

        lsv = LanguageSelectorViewlet(context, self.request, None, None)
        translations = lsv._translations(missing)

        # We want to see the babel_view
        append_path = (
            "",
            "babel_view",
        )

        non_viewable = set()
        for data in results:
            code = str(data["code"])
            data["translated"] = code in translations.keys()

            appendtourl = "/".join(append_path)

            if data["translated"]:
                trans, direct, has_view_permission = translations[code]
                if not has_view_permission:
                    # shortcut if the user cannot see the item
                    non_viewable.add(data["code"])
                    continue
                data["url"] = trans.absolute_url() + appendtourl
            else:
                non_viewable.add(data["code"])

        # filter out non-viewable items
        results = [r for r in results if r["code"] not in non_viewable]
        return results

    def portal_url(self):
        portal_tool = getToolByName(self.context, "portal_url", None)
        if portal_tool is not None:
            return portal_tool.getPortalObject().absolute_url()
        return None

    def render(self):
        self.request["disable_border"] = True
        self.request["disable_plone.leftcolumn"] = True
        self.request["disable_plone.rightcolumn"] = True

        for field in self.fields.keys():
            if field in self.schema:
                if ILanguageIndependentField.providedBy(self.schema[field]):
                    self.widgets[field].addClass("languageindependent")
            # With plone.autoform, fieldnames from additional schematas
            # reference their schema by prefixing their fieldname
            # with schema.__identifier__ and then a dot as a separator
            # See autoform.txt in the autoform package
            if "." in field:
                schemaname, fieldname = field.split(".")
                for schema in self.additionalSchemata:
                    if schemaname == schema.__identifier__ and fieldname in schema:
                        if ILanguageIndependentField.providedBy(
                            schema[fieldname]
                        ):  # noqa
                            self.widgets[field].addClass("languageindependent")
        self.babel_content = super().render()
        return self.babel()

    @property
    def max_nr_of_buttons(self):
        registry = getUtility(IRegistry)
        settings = registry.forInterface(
            IMultiLanguageExtraOptionsSchema, prefix="plone"
        )
        return settings.buttons_babel_view_up_to_nr_translations


DefaultMultilingualEditView = layout.wrap_form(MultilingualEditForm)



from plone.app.i18n.locales.browser.selector import LanguageSelector
from plone.app.multilingual import api
from plone.app.multilingual.browser.helper_views import selector_view
from plone.app.multilingual.browser.selector import addQuery
from plone.app.multilingual.browser.selector import getPostPath
from plone.app.multilingual.browser.selector import LanguageSelectorViewlet
from plone.app.multilingual.browser.selector import NOT_TRANSLATED_YET_TEMPLATE
from plone.app.multilingual.interfaces import IMultiLanguageExtraOptionsSchema
from plone.app.multilingual.interfaces import IPloneAppMultilingualInstalled
from plone.app.multilingual.interfaces import ITG
from plone.app.multilingual.testing import PAM_FUNCTIONAL_TESTING
from plone.app.multilingual.testing import PAM_INTEGRATION_PRESET_TESTING
from plone.app.multilingual.testing import PAM_INTEGRATION_TESTING
from plone.dexterity.utils import createContentInContainer
from plone.i18n.interfaces import ILanguageSchema
from plone.registry.interfaces import IRegistry
from plone.testing.z2 import Browser
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.utils import safe_nativestring
from urllib.parse import parse_qs
from urllib.parse import urlparse
from zope.component import getUtility
from zope.event import notify
from zope.interface import alsoProvides
from zope.lifecycleevent import ObjectModifiedEvent

import json
import lxml
import transaction
import unittest


class EvilObject:
    def __str__(self):
        raise UnicodeError

    def __unicode__(self):
        raise UnicodeError


SELECTOR_VIEW_TEMPLATE = "%(url)s/@@multilingual-selector/%(tg)s/%(lang)s%(query)s"


class TestLanguageSelectorBasics(unittest.TestCase):
    layer = PAM_FUNCTIONAL_TESTING

    def setUp(self):
        # Set test variables
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        self.request = self.layer["request"]
        alsoProvides(self.layer["request"], IPloneAppMultilingualInstalled)

        # Setup testbrowser
        self.browser = Browser(self.layer["app"])
        self.browser.handleErrors = False

    def test_selector_viewlet_is_available(self):
        selector_viewlet = LanguageSelectorViewlet(
            self.portal, self.request, None, None
        )
        selector_viewlet.update()
        self.assertTrue(selector_viewlet.available())

    def test_selector_viewlet_is_available_but_tool_is_not(self):
        selector_viewlet = LanguageSelectorViewlet(
            self.portal, self.request, None, None
        )
        selector_viewlet.update()
        selector_viewlet.tool = None
        self.assertFalse(selector_viewlet.available())

    def test_selector_data_not_modified_by_viewlet(self):
        selector_viewlet = LanguageSelectorViewlet(
            self.portal, self.request, None, None
        )
        selector_viewlet.update()
        selector_viewlet_languages = selector_viewlet.languages()

        selector_adapter = LanguageSelector(self.portal, self.request, None, None)
        selector_adapter.update()
        selector_adapter_languages = selector_adapter.languages()

        self.assertNotEqual(selector_adapter_languages, selector_viewlet_languages)

    def assertFullyTranslatedPages(self):
        a = createContentInContainer(
            self.portal["en"], "Document", title="Test document"
        )

        a_ca = api.translate(a, "ca")
        a_ca.setTitle("Test Document (CA)")

        a_es = api.translate(a, "es")
        a_es.setTitle("Test Document (ES)")

        wftool = getToolByName(self.portal, "portal_workflow")
        wftool.doActionFor(a, "publish")
        wftool.doActionFor(a_ca, "publish")
        wftool.doActionFor(a_es, "publish")

        notify(ObjectModifiedEvent(a))
        notify(ObjectModifiedEvent(a_ca))
        notify(ObjectModifiedEvent(a_es))

        selector_viewlet = LanguageSelectorViewlet(a, self.request, None, None)
        selector_viewlet.update()
        selector_viewlet_languages = selector_viewlet.languages()

        expected_languages = [
            {
                "code": "en",
                "flag": "countryflag/gb",
                "name": "English",
                "native": "English",
                "url": SELECTOR_VIEW_TEMPLATE
                % {
                    "url": self.portal_url,
                    "tg": ITG(a),
                    "lang": "en",
                    "query": "?set_language=en",
                },
                "selected": True,
                "translated": True,
            },
            {
                "code": "ca",
                "flag": "languageflag/ca",
                "name": "Catalan",
                "native": "Catal\xe0",
                "url": SELECTOR_VIEW_TEMPLATE
                % {
                    "url": self.portal_url,
                    "tg": ITG(a),
                    "lang": "ca",
                    "query": "?set_language=ca",
                },
                "selected": False,
                "translated": True,
            },
            {
                "code": "es",
                "flag": "countryflag/es",
                "name": "Spanish",
                "native": "Espa\xf1ol",
                "url": SELECTOR_VIEW_TEMPLATE
                % {
                    "url": self.portal_url,
                    "tg": ITG(a),
                    "lang": "es",
                    "query": "?set_language=es",
                },
                "selected": False,
                "translated": True,
            },
        ]

        self.assertEqual(
            json.dumps(selector_viewlet_languages, sort_keys=True),
            json.dumps(expected_languages, sort_keys=True),
        )

        transaction.commit()

        self.browser.open(selector_viewlet_languages[0]["url"])
        self.assertEqual(self.browser.url, a.absolute_url() + "?set_language=en")
        self.assertRegex(self.browser.contents, r"Distributed under the")

        self.browser.open(selector_viewlet_languages[1]["url"])
        self.assertEqual(self.browser.url, a_ca.absolute_url() + "?set_language=ca")

        self.assertIn('lang="ca"', self.browser.contents)

        self.browser.open(selector_viewlet_languages[2]["url"])
        self.assertEqual(self.browser.url, a_es.absolute_url() + "?set_language=es")
        self.assertIn('lang="es"', self.browser.contents)

    def test_languages_fully_translated_by_closest(self):
        self.registry = getUtility(IRegistry)
        self.settings = self.registry.forInterface(
            IMultiLanguageExtraOptionsSchema, prefix="plone"
        )
        self.settings.selector_lookup_translations_policy = "closest"

        self.assertFullyTranslatedPages()

    def test_languages_fully_translated_by_dialog(self):
        self.registry = getUtility(IRegistry)
        self.settings = self.registry.forInterface(
            IMultiLanguageExtraOptionsSchema, prefix="plone"
        )
        self.settings.selector_lookup_translations_policy = "dialog"

        self.assertFullyTranslatedPages()

    def setUpPartiallyTranslatedContent(self):
        wftool = getToolByName(self.portal, "portal_workflow")

        f_en = createContentInContainer(
            self.portal["en"], "Folder", title="Test folder"
        )
        wftool.doActionFor(f_en, "publish")

        a_en = createContentInContainer(
            self.portal["en"]["test-folder"],
            "Document",
            title="Test document",
        )
        wftool.doActionFor(a_en, "publish")

        f_ca = api.translate(f_en, "ca")
        f_ca.setTitle("Test folder CA")
        wftool.doActionFor(f_ca, "publish")

        transaction.commit()

    def setUpFullyTranslatedContent(self):
        wftool = getToolByName(self.portal, "portal_workflow")

        self.setUpPartiallyTranslatedContent()

        a_ca = api.translate(self.portal["en"]["test-folder"]["test-document"], "ca")
        a_ca.setTitle("Test document CA")
        wftool.doActionFor(a_ca, "publish")

        f_es = api.translate(self.portal["en"]["test-folder"], "es")
        f_es.setTitle("Test folder ES")
        wftool.doActionFor(f_es, "publish")

        a_es = api.translate(self.portal["en"]["test-folder"]["test-document"], "es")
        a_es.setTitle("Test document ES")
        wftool.doActionFor(a_es, "publish")

        transaction.commit()

    def test_languages_untranslated_by_closest(self):
        # Define selector policy
        self.registry = getUtility(IRegistry)
        self.settings = self.registry.forInterface(
            IMultiLanguageExtraOptionsSchema, prefix="plone"
        )
        self.settings.selector_lookup_translations_policy = "closest"

        wftool = getToolByName(self.portal, "portal_workflow")

        a_ca = createContentInContainer(
            self.portal["ca"], "Document", title="Test document"
        )
        a_ca.setLanguage("ca")
        wftool.doActionFor(a_ca, "publish")

        view = selector_view(a_ca, self.request)
        view.tg = ITG(a_ca)

        # Check that the 'original' language gets the created document
        view.lang = "ca"
        url = view.getClosestDestination()
        self.assertEqual(url, a_ca.absolute_url())

        # Check that untranslated languages get the nearest translated parent
        view.lang = "es"
        url = view.getClosestDestination()
        self.assertEqual(url, self.portal["es"].absolute_url())

        view.lang = "en"
        url = view.getClosestDestination()
        self.assertEqual(url, self.portal["en"].absolute_url())

        # Check that translating content updates the closest destination
        a_es = api.translate(a_ca, "es")
        view.lang = "es"
        url = view.getClosestDestination()
        self.assertEqual(url, a_es.absolute_url())

    def test_languages_partially_translated_by_closest(self):
        self.registry = getUtility(IRegistry)
        self.settings = self.registry.forInterface(
            IMultiLanguageExtraOptionsSchema, prefix="plone"
        )
        self.settings.selector_lookup_translations_policy = "closest"

        # Setup content
        self.setUpPartiallyTranslatedContent()

        a_en = self.portal["en"]["test-folder"]["test-document"]
        f_ca = self.portal["ca"]["test-folder"]

        selector = LanguageSelectorViewlet(a_en, self.request, None, None)
        selector.update()
        selector_languages = selector.languages()

        # Check that EN translation is found
        self.browser.open(selector_languages[0]["url"])
        self.assertEqual(self.browser.url, a_en.absolute_url() + "?set_language=en")
        self.assertIn('lang="en"', self.browser.contents)
        # But extra check, because English is the default?
        self.assertRegex(self.browser.contents, r"Distributed under the")

        # Check that CA translation is only partial and a parent folder
        # is found
        self.browser.open(selector_languages[1]["url"])
        self.assertEqual(self.browser.url, f_ca.absolute_url() + "?set_language=ca")
        self.assertIn('lang="ca"', self.browser.contents)

        # Check that ES translation is missing and only the LRF is found
        self.browser.open(selector_languages[2]["url"])
        self.assertEqual(
            self.browser.url,
            self.portal["es"].absolute_url() + "?set_language=es",
        )
        self.assertIn('lang="es"', self.browser.contents)

    def test_languages_partially_translated_by_dialog(self):
        self.registry = getUtility(IRegistry)
        self.settings = self.registry.forInterface(
            IMultiLanguageExtraOptionsSchema, prefix="plone"
        )
        self.settings.selector_lookup_translations_policy = "dialog"

        # Setup content
        self.setUpPartiallyTranslatedContent()

        a_en = self.portal["en"]["test-folder"]["test-document"]

        selector = LanguageSelectorViewlet(a_en, self.request, None, None)
        selector.update()
        selector_languages = selector.languages()

        # Check that EN translation is found
        self.browser.open(selector_languages[0]["url"])
        self.assertEqual(self.browser.url, a_en.absolute_url() + "?set_language=en")
        # But extra check, because English is the default?
        self.assertIn('lang="en"', self.browser.contents)
        self.assertRegex(self.browser.contents, r"Distributed under the")

        tgid = selector_languages[1]["url"].split("/")[-2]

        # Check that CA is not yet translated
        self.browser.open(selector_languages[1]["url"])
        self.assertEqual(
            self.browser.url,
            self.portal.absolute_url()
            + NOT_TRANSLATED_YET_TEMPLATE
            + "/"
            + tgid
            + "?set_language=ca",
        )
        self.assertIn('lang="ca"', self.browser.contents)

        # Check that ES is not yet translated
        self.browser.open(selector_languages[2]["url"])
        self.assertEqual(
            self.browser.url,
            self.portal.absolute_url()
            + NOT_TRANSLATED_YET_TEMPLATE
            + "/"
            + tgid
            + "?set_language=es",
        )
        self.assertIn('lang="es"', self.browser.contents)

    def assertSiteRoot(self):
        self.selector_viewlet = LanguageSelectorViewlet(
            self.portal, self.request, None, None
        )
        self.selector_viewlet.update()
        selector_languages = self.selector_viewlet.languages()

        transaction.commit()

        # Check EN root
        self.assertEqual(
            selector_languages[0]["url"],
            (
                self.portal.absolute_url()
                + "/@@multilingual-selector/notg/en?set_language=en"
            ),
        )
        self.browser.open(selector_languages[0]["url"])
        self.assertIn('lang="en"', self.browser.contents)
        self.assertRegex(self.browser.contents, r"Distributed under the")

        # Check CA root
        self.assertEqual(
            selector_languages[1]["url"],
            (
                self.portal.absolute_url()
                + "/@@multilingual-selector/notg/ca?set_language=ca"
            ),
        )
        self.browser.open(selector_languages[1]["url"])
        self.assertIn('lang="ca"', self.browser.contents)

        # Check ES root
        self.browser.open(selector_languages[2]["url"])
        self.assertEqual(
            selector_languages[2]["url"],
            (
                self.portal.absolute_url()
                + "/@@multilingual-selector/notg/es?set_language=es"
            ),
        )
        self.assertIn('lang="es"', self.browser.contents)

    def test_siteroot_by_closest(self):
        self.registry = getUtility(IRegistry)
        self.settings = self.registry.forInterface(
            IMultiLanguageExtraOptionsSchema, prefix="plone"
        )
        self.settings.selector_lookup_translations_policy = "closest"

        self.assertSiteRoot()

    def test_siteroot_by_dialog(self):
        self.registry = getUtility(IRegistry)
        self.settings = self.registry.forInterface(
            IMultiLanguageExtraOptionsSchema, prefix="plone"
        )
        self.settings.selector_lookup_translations_policy = "dialog"

        self.assertSiteRoot()

    def assertRootFolders(self):
        self.selector_viewlet = LanguageSelectorViewlet(
            self.portal["en"], self.request, None, None
        )
        self.selector_viewlet.update()

        selector_languages = self.selector_viewlet.languages()
        tg = ITG(self.portal["en"])

        expected_languages = [
            {
                "code": "en",
                "flag": "countryflag/gb",
                "name": "English",
                "native": "English",
                "url": SELECTOR_VIEW_TEMPLATE
                % {
                    "url": self.portal.absolute_url(),
                    "tg": tg,
                    "lang": "en",
                    "query": "?set_language=en",
                },
                "selected": True,
                "translated": True,
            },
            {
                "code": "ca",
                "flag": "languageflag/ca",
                "name": "Catalan",
                "native": "Catal\xe0",
                "url": SELECTOR_VIEW_TEMPLATE
                % {
                    "url": self.portal.absolute_url(),
                    "tg": tg,
                    "lang": "ca",
                    "query": "?set_language=ca",
                },
                "selected": False,
                "translated": True,
            },
            {
                "code": "es",
                "flag": "countryflag/es",
                "name": "Spanish",
                "native": "Espa\xf1ol",
                "url": SELECTOR_VIEW_TEMPLATE
                % {
                    "url": self.portal.absolute_url(),
                    "tg": tg,
                    "lang": "es",
                    "query": "?set_language=es",
                },
                "selected": False,
                "translated": True,
            },
        ]

        self.assertEqual(
            json.dumps(selector_languages, sort_keys=True),
            json.dumps(expected_languages, sort_keys=True),
        )

        # Check EN
        self.browser.open(selector_languages[0]["url"])
        self.assertEqual(
            self.browser.url,
            self.portal["en"].absolute_url() + "?set_language=en",
        )
        self.assertIn('lang="en"', self.browser.contents)
        self.assertRegex(self.browser.contents, r"Distributed under the")

        # Check CA
        self.browser.open(selector_languages[1]["url"])
        self.assertEqual(
            self.browser.url,
            self.portal.ca.absolute_url() + "?set_language=ca",
        )
        self.assertIn('lang="ca"', self.browser.contents)

        # Check ES
        self.browser.open(selector_languages[2]["url"])
        self.assertEqual(
            self.browser.url,
            self.portal.es.absolute_url() + "?set_language=es",
        )
        self.assertIn('lang="es"', self.browser.contents)

    def test_languages_root_folders_by_dialog(self):
        self.registry = getUtility(IRegistry)
        self.settings = self.registry.forInterface(
            IMultiLanguageExtraOptionsSchema, prefix="plone"
        )
        self.settings.selector_lookup_translations_policy = "dialog"

        self.assertRootFolders()

    def test_languages_root_folders_by_closest(self):
        self.registry = getUtility(IRegistry)
        self.settings = self.registry.forInterface(
            IMultiLanguageExtraOptionsSchema, prefix="plone"
        )
        self.settings.selector_lookup_translations_policy = "closest"

        self.assertRootFolders()

    def test_languages_preserve_view(self):
        self.registry = getUtility(IRegistry)
        self.settings = self.registry.forInterface(
            IMultiLanguageExtraOptionsSchema, prefix="plone"
        )
        self.settings.selector_lookup_translations_policy = "closest"

        self.setUpPartiallyTranslatedContent()

        f_en = self.portal["en"]["test-folder"]
        f_en = self.portal["en"]["test-folder"]
        f_ca = self.portal["ca"]["test-folder"]

        self.request["PATH_INFO"] = "/plone/en/test-folder/contact-info"

        selector = LanguageSelectorViewlet(f_en, self.request, None, None)
        selector.update()
        selector_languages = selector.languages()

        transaction.commit()

        # Check EN
        self.browser.open(selector_languages[0]["url"])
        self.assertEqual(
            self.browser.url,
            f_en.absolute_url() + "/contact-info?set_language=en",
        )
        self.assertIn('lang="en"', self.browser.contents)
        self.assertRegex(self.browser.contents, r"Distributed under the")

        # Check CA
        self.browser.open(selector_languages[1]["url"])
        self.assertEqual(
            self.browser.url,
            f_ca.absolute_url() + "/contact-info?set_language=ca",
        )
        self.assertIn('lang="ca"', self.browser.contents)

        # Check ES
        self.browser.open(selector_languages[2]["url"])
        # Here @@search isn't preserved because we've gone up a level
        self.assertEqual(
            self.browser.url,
            self.portal.es.absolute_url() + "?set_language=es",
        )
        self.assertIn('lang="es"', self.browser.contents)

        # Shouldn't do for the not-translated-yet thingie when dialog is on
        self.registry = getUtility(IRegistry)
        self.settings = self.registry.forInterface(
            IMultiLanguageExtraOptionsSchema, prefix="plone"
        )
        self.settings.selector_lookup_translations_policy = "dialog"

        selector = LanguageSelectorViewlet(f_en, self.request, None, None)
        selector.update()
        selector_languages = selector.languages()

        transaction.commit()

        # Check EN
        self.browser.open(selector_languages[0]["url"])
        self.assertEqual(
            self.browser.url,
            f_en.absolute_url() + "/contact-info?set_language=en",
        )
        self.assertIn('lang="en"', self.browser.contents)
        self.assertRegex(self.browser.contents, r"Distributed under the")

        # Check CA
        self.browser.open(selector_languages[1]["url"])
        self.assertEqual(
            self.browser.url,
            f_ca.absolute_url() + "/contact-info?set_language=ca",
        )
        self.assertIn('lang="ca"', self.browser.contents)

        # Check ES
        self.browser.open(selector_languages[2]["url"])
        # Here @@search isn't preserved because we've got the dialog
        tgid = selector_languages[2]["url"].split("/")[-3]
        self.assertEqual(
            self.browser.url,
            self.portal.absolute_url()
            + NOT_TRANSLATED_YET_TEMPLATE
            + "/"
            + tgid
            + "?set_language=es",
        )
        self.assertIn('lang="es"', self.browser.contents)

    def assertLanguagesPreserveQuery(self, policy):
        self.setUpPartiallyTranslatedContent()
        self.request["PATH_INFO"] = "/plone/en/test-folder/contact-info"
        self.request.form["uni"] = "pres\xd8rved"
        self.request.form["int"] = "1"

        f_en = self.portal["en"]["test-folder"]
        f_ca = self.portal["ca"]["test-folder"]

        selector = LanguageSelectorViewlet(f_en, self.request, None, None)
        selector.update()
        selector_languages = selector.languages()

        tgid = selector_languages[2]["url"].split("/")[-3]
        untranslated_url = {
            "closest": (
                self.portal["es"].absolute_url()
                + "?int=1&uni=pres%C3%98rved&set_language=es"
            ),
            "dialog": (
                self.portal.absolute_url()
                + NOT_TRANSLATED_YET_TEMPLATE
                + "/"
                + tgid
                + "?int=1&uni=pres%C3%98rved&set_language=es"
            ),
        }

        transaction.commit()

        # Check EN
        self.browser.open(selector_languages[0]["url"])
        self.assertUrlsEqual(
            self.browser.url,
            (
                f_en.absolute_url()
                + "/contact-info?int=1&uni=pres%C3%98rved&set_language=en"
            ),
        )
        self.assertIn('lang="en"', self.browser.contents)
        self.assertRegex(self.browser.contents, r"Distributed under the")

        # Check CA
        self.browser.open(selector_languages[1]["url"])
        self.assertUrlsEqual(
            self.browser.url,
            (
                f_ca.absolute_url()
                + "/contact-info?int=1&uni=pres%C3%98rved&set_language=ca"
            ),
        )
        self.assertIn('lang="ca"', self.browser.contents)

        # Check ES
        self.browser.open(selector_languages[2]["url"])
        # Here @@search isn't preserved because we've got the dialog
        url = self.browser.url
        self.assertUrlsEqual(url, untranslated_url[policy])
        self.assertIn('lang="es"', self.browser.contents)

    def test_languages_preserve_by_closest(self):
        self.registry = getUtility(IRegistry)
        self.settings = self.registry.forInterface(
            IMultiLanguageExtraOptionsSchema, prefix="plone"
        )
        self.settings.selector_lookup_translations_policy = "closest"

        self.assertLanguagesPreserveQuery("closest")

    def test_languages_preserve_by_dialog(self):
        self.registry = getUtility(IRegistry)
        self.settings = self.registry.forInterface(
            IMultiLanguageExtraOptionsSchema, prefix="plone"
        )
        self.settings.selector_lookup_translations_policy = "dialog"

        self.assertLanguagesPreserveQuery("dialog")

    # XXX: this seems to me like a demented use case.
    # If you put the VH _after_ the jail,
    # for example mapping mysite.cat to /plone/ca and mysite.com to /plone/en,
    # then I can't possibly make the whole language switcher work
    # because I have no idea to which domain I should reroute the request
    # (that information is sitting in nginx|apache config,
    # therefore I cannot conceivably gather it).
    # Therefore, while PAM knows which is the translation object,
    # it can't figure out its URL correctly.
    # At most, it can leverage acquisition so that,
    # if you switch from mysite.com/my/object to catalan,
    # it goes to mysite.com/ca/my/object
    # (which has a path of /plone/en/my/object/ca/my/object),
    # but that is too fragile and sucky for me to take care of supporting it.
    #
    # The original test case is below: if you wrote it,
    # I'll be glad to hear which was the use case
    # and what is your idea to make it work.
    #
    # def test_languages_vhr(self):
    #     registry = getUtility(IRegistry)
    #     settings = registry.forInterface(IMultiLanguageExtraOptionsSchema)
    #     settings.selector_lookup_translations_policy = 'dialog'
    #
    #     from Products.CMFCore.interfaces import ISiteRoot
    #     from zope.interface import alsoProvides
    #     provideAdapter(DummyState, adapts=(Dummy, DummyRequest),
    #                    provides=Interface, name="plone_context_state")
    #     context = Dummy()
    #     context.__parent__ = self.portal
    #     context.portal_url = Dummy()
    #     container = Dummy()
    #     context = context.__of__(container)
    #     alsoProvides(container, ISiteRoot)
    #     request = DummyRequest()
    #     selector = LanguageSelectorViewlet(context, request, None, None)
    #     context.physicalpath = ['', 'fake', 'path']
    #     vbase = '/VirtualHostBase/http/127.0.0.1/'
    #     request.PATH_INFO = vbase + 'fake/path/VirtualHostRoot/to/object'
    #     request.form['uni'] = u'pres\xd8rved'
    #     request.form['int'] = '1'
    #     notify(ObjectCreatedEvent(context))
    #     IUUID(context, None)
    #     selector.update()
    #     selector.tool = MockLanguageTool()
    #     base = 'object_url/to/object?int=1&uni='
    #     expected = [
    #         {'code': 'nl',
    #          'translated': True,
    #          'selected': False,
    #          'url': base + 'pres%C3%98rved&set_language=nl'},
    #         {'code': 'en',
    #          'translated': True,
    #          'selected': True,
    #          'url': base + 'pres%C3%98rved&set_language=en'},
    #         {'code': 'no',
    #          'translated': False,
    #          'selected': False,
    #          'url': base + 'pres%C3%98rved&set_language=no'}]
    #     self.assertEqual(selector.languages(), expected)

    def assertUrlsEqual(self, url1, url2):
        """The order of query-strings is sometimes random in python 3
        This compares urls disregarding the order.
        """
        parsed_url_1 = urlparse(safe_nativestring(url1))
        parse_qs_1 = parse_qs(parsed_url_1.query)
        parsed_url_2 = urlparse(safe_nativestring(url2))
        parse_qs_2 = parse_qs(parsed_url_2.query)
        self.assertEqual(parsed_url_1[0], parsed_url_2[0])
        self.assertEqual(parsed_url_1[1], parsed_url_2[1])
        self.assertEqual(parsed_url_1[2], parsed_url_2[2])
        self.assertEqual(parsed_url_1[3], parsed_url_2[3])
        self.assertEqual(parse_qs_1, parse_qs_2)


class TestLanguageSelectorPostPath(unittest.TestCase):
    layer = PAM_FUNCTIONAL_TESTING

    def setUp(self):
        alsoProvides(self.layer["request"], IPloneAppMultilingualInstalled)
        self.portal = self.layer["portal"]
        self.folder = createContentInContainer(
            self.portal["ca"], "Folder", title="Folder"
        )
        self.document = createContentInContainer(
            self.portal["ca"], "Document", title="Document"
        )
        self.request = self.layer["request"]

    def test_findpath(self):
        self.request["PATH_INFO"] = "/plone/ca/folder/document/whatever"
        self.assertEqual(getPostPath(self.document, self.request), "/whatever")

    def test_findpath_match(self):
        self.request["PATH_INFO"] = "/plone/ca/folder/document"
        self.assertEqual(getPostPath(self.document, self.request), "")

    def test_findpath_match_slash(self):
        self.request["PATH_INFO"] = "/plone/ca/folder/document/"
        self.assertEqual(getPostPath(self.document, self.request), "")

    def test_findpath_template(self):
        self.request["PATH_INFO"] = "/plone/ca/folder/document/" "whatever/foo/edit"
        self.assertEqual(getPostPath(self.document, self.request), "/whatever/foo/edit")

    def test_findpath_view(self):
        self.request["PATH_INFO"] = "/plone/ca/folder/document/contact-info"
        self.assertEqual(getPostPath(self.document, self.request), "/contact-info")

    def test_findpath_vhr(self):
        self.request["PATH_INFO"] = (
            "/VirtualHostBase/http/127.0.0.1"
            "/plone/ca/folder/document/"
            "VirtualHostRoot/whatever"
        )
        self.assertEqual(getPostPath(self.document, self.request), "/whatever")

    def test_findpath_vh_marker(self):
        self.request["PATH_INFO"] = (
            "/VirtualHostBase/http/127.0.0.1"
            "/plone/ca/folder/document/"
            "VirtualHostRoot/_vh_foo/whatever"
        )
        self.assertEqual(getPostPath(self.document, self.request), "/whatever")

    def test_findpath_vhr_and_traverser(self):
        self.request["PATH_INFO"] = (
            "/VirtualHostBase/http/127.0.0.1"
            "/plone/ca/folder/document/++theme++foo/"
            "VirtualHostRoot/whatever"
        )
        self.assertEqual(getPostPath(self.document, self.request), "/whatever")


class TestLanguageSelectorAddQuery(unittest.TestCase):
    layer = PAM_INTEGRATION_TESTING

    def setUp(self):
        self.request = self.layer["request"]
        alsoProvides(self.layer["request"], IPloneAppMultilingualInstalled)
        self.url = "/foo/bar/"

    def test_formvariables(self):
        self.request.form["one"] = 1
        self.request.form["two"] = 2
        self.assertUrlsEqual(
            addQuery(self.request, self.url), self.url + "?two:int=2&one:int=1"
        )

    def test_formvariables_sequences(self):
        self.request.form["one"] = ("a",)
        self.request.form["two"] = ["b", 2]
        self.assertUrlsEqual(
            addQuery(self.request, self.url),
            self.url + "?two:list=b&two:int:list=2&one=%28%27a%27%2C%29",
        )

    def test_formvariables_unicode(self):
        self.request.form["one"] = "Før"
        self.request.form["two"] = "foo"
        self.assertUrlsEqual(
            addQuery(self.request, self.url),
            self.url + "?two=foo&one=F%C3%B8r",
        )

    def test_formvariables_utf8(self):
        self.request.form["one"] = "Før"
        self.request.form["two"] = "foo"
        self.assertUrlsEqual(
            addQuery(self.request, self.url),
            self.url + "?two=foo&one=F%C3%B8r",
        )

    def test_formvariables_object(self):
        self.request.form["one"] = "1"
        self.request.form["two"] = EvilObject()
        self.assertUrlsEqual(addQuery(self.request, self.url), self.url)

    def test_formvariables_exclude(self):
        self.request.form["one"] = 1
        self.request.form["two"] = 2
        self.assertUrlsEqual(
            addQuery(self.request, self.url, exclude=("two",)),
            self.url + "?one:int=1",
        )

    def test_formvariables_extras(self):
        self.request.form["one"] = 1
        self.request.form["two"] = 2
        self.assertUrlsEqual(
            addQuery(self.request, self.url, three=3),
            self.url + "?one:int=1&three:int=3&two:int=2",
        )

    def assertUrlsEqual(self, url1, url2):
        """The order of query-strings is sometimes random in python 3
        This compares urls disregarding the order.
        """
        parsed_url_1 = urlparse(url1)
        parse_qs_1 = parse_qs(parsed_url_1.query)
        parsed_url_2 = urlparse(url2)
        parse_qs_2 = parse_qs(parsed_url_2.query)
        self.assertEqual(parsed_url_1[0], parsed_url_2[0])
        self.assertEqual(parsed_url_1[1], parsed_url_2[1])
        self.assertEqual(parsed_url_1[2], parsed_url_2[2])
        self.assertEqual(parsed_url_1[3], parsed_url_2[3])
        self.assertEqual(parse_qs_1, parse_qs_2)


class TestLanguageSelectorSetLanguage(unittest.TestCase):
    layer = PAM_INTEGRATION_PRESET_TESTING

    def setUp(self):
        # Set test variables
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        self.request = self.layer["request"]
        alsoProvides(self.layer["request"], IPloneAppMultilingualInstalled)

        registry = getUtility(IRegistry)
        self.settings = registry.forInterface(ILanguageSchema, prefix="plone")

    def test_set_language_is_present(self):
        """test the presence of set_language parameter in the urls created in the language selector"""

        self.settings.set_cookie_always = False
        self.selector_viewlet = LanguageSelectorViewlet(
            self.portal["en"], self.request, None, None
        )
        self.selector_viewlet.update()

        selector_languages = self.selector_viewlet.languages()
        tg = ITG(self.portal["en"])

        expected_languages = [
            {
                "code": "en",
                "flag": "countryflag/gb",
                "name": "English",
                "native": "English",
                "url": SELECTOR_VIEW_TEMPLATE
                % {
                    "url": self.portal.absolute_url(),
                    "tg": tg,
                    "lang": "en",
                    "query": "?set_language=en",
                },
                "selected": True,
                "translated": True,
            },
            {
                "code": "ca",
                "flag": "languageflag/ca",
                "name": "Catalan",
                "native": "Catal\xe0",
                "url": SELECTOR_VIEW_TEMPLATE
                % {
                    "url": self.portal.absolute_url(),
                    "tg": tg,
                    "lang": "ca",
                    "query": "?set_language=ca",
                },
                "selected": False,
                "translated": True,
            },
            {
                "code": "es",
                "flag": "countryflag/es",
                "name": "Spanish",
                "native": "Espa\xf1ol",
                "url": SELECTOR_VIEW_TEMPLATE
                % {
                    "url": self.portal.absolute_url(),
                    "tg": tg,
                    "lang": "es",
                    "query": "?set_language=es",
                },
                "selected": False,
                "translated": True,
            },
        ]
        self.assertEqual(
            json.dumps(selector_languages, sort_keys=True),
            json.dumps(expected_languages, sort_keys=True),
        )

    def test_set_language_is_not_present_when_always_set_cookie_is_set(self):
        """test the absence of set_language parameter in the urls created in the language selector"""
        self.settings.set_cookie_always = True
        self.selector_viewlet = LanguageSelectorViewlet(
            self.portal["en"], self.request, None, None
        )
        self.selector_viewlet.update()

        selector_languages = self.selector_viewlet.languages()
        tg = ITG(self.portal["en"])

        expected_languages = [
            {
                "code": "en",
                "flag": "countryflag/gb",
                "name": "English",
                "native": "English",
                "url": SELECTOR_VIEW_TEMPLATE
                % {
                    "url": self.portal.absolute_url(),
                    "tg": tg,
                    "lang": "en",
                    "query": "",
                },
                "selected": True,
                "translated": True,
            },
            {
                "code": "ca",
                "flag": "languageflag/ca",
                "name": "Catalan",
                "native": "Catal\xe0",
                "url": SELECTOR_VIEW_TEMPLATE
                % {
                    "url": self.portal.absolute_url(),
                    "tg": tg,
                    "lang": "ca",
                    "query": "",
                },
                "selected": False,
                "translated": True,
            },
            {
                "code": "es",
                "flag": "countryflag/es",
                "name": "Spanish",
                "native": "Espa\xf1ol",
                "url": SELECTOR_VIEW_TEMPLATE
                % {
                    "url": self.portal.absolute_url(),
                    "tg": tg,
                    "lang": "es",
                    "query": "",
                },
                "selected": False,
                "translated": True,
            },
        ]

        self.assertEqual(
            json.dumps(selector_languages, sort_keys=True),
            json.dumps(expected_languages, sort_keys=True),
        )


class TestLanguageSelectorDisplayOptions(unittest.TestCase):
    layer = PAM_FUNCTIONAL_TESTING

    def setUp(self):
        # Set test variables
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        self.request = self.layer["request"]
        alsoProvides(self.layer["request"], IPloneAppMultilingualInstalled)

        # Setup testbrowser
        self.browser = Browser(self.layer["app"])
        self.browser.handleErrors = False

    def test_language_selector_flag_is_a_svg(self):
        registry = getUtility(IRegistry)
        self.settings = registry.forInterface(ILanguageSchema, prefix="plone")
        self.settings.display_flags = True
        self.settings.always_show_selector = True

        transaction.commit()

        self.browser.open(self.portal_url)
        output = lxml.html.fromstring(self.browser.contents)
        svgs = output.xpath('//svg[contains(@class, "plone-icon-flag")]')
        self.assertGreater(len(svgs), 0)



from zope.interface import Attribute
from zope.interface import implementer
from zope.interface.interfaces import IObjectEvent


class IObjectWillBeTranslatedEvent(IObjectEvent):
    """Sent before an object is translated."""

    object = Attribute("The object to be translated.")
    language = Attribute("Target language.")


class IObjectTranslatedEvent(IObjectEvent):
    """Sent after an object was translated."""

    object = Attribute("The object to be translated.")
    target = Attribute("The translation target object.")
    language = Attribute("Target language.")


class ITranslationRegisteredEvent(IObjectEvent):
    """Sent after a new translation was registered.

    This is meant to be notified on low-level manager level only.
    """

    object = Attribute("The base object.")
    target = Attribute("The translated object.")
    language = Attribute("The language of the translated obejct.")


class ITranslationUpdatedEvent(IObjectEvent):
    """Sent after an translation was moved to point to a different object

    This is meant to be notified on low-level manager level only.
    """

    object = Attribute("The base object.")
    old_object = Attribute("The old translation, now orphaned.")
    language = Attribute("The language of the objects.")


class ITranslationRemovedEvent(IObjectEvent):
    """Sent after an translation was removed.

    This is meant to be notified on low-level manager level only.
    """

    object = Attribute("The base object.")
    old_object = Attribute("The old translation, now orphaned.")
    language = Attribute("The language of the objects.")


@implementer(IObjectWillBeTranslatedEvent)
class ObjectWillBeTranslatedEvent:
    """Sent before an object is translated."""

    def __init__(self, context, language):
        self.object = context
        self.language = language


@implementer(IObjectTranslatedEvent)
class ObjectTranslatedEvent:
    """Sent after an object was translated."""

    def __init__(self, context, target, language):
        self.object = context
        self.target = target
        self.language = language


@implementer(ITranslationRegisteredEvent)
class TranslationRegisteredEvent:
    """Sent after a new translation was registered."""

    def __init__(self, context, target, language):
        self.object = context
        self.target = target
        self.language = language


@implementer(ITranslationUpdatedEvent)
class TranslationUpdatedEvent:
    """Sent after an translation was moved to point to a different object"""

    def __init__(self, context, old_object, language):
        self.object = context
        self.old_object = old_object
        self.language = language


@implementer(ITranslationRemovedEvent)
class TranslationRemovedEvent:
    """Sent after an translation was moved to point to a different object"""

    def __init__(self, context, old_object, language):
        self.object = context
        self.old_object = old_object
        self.language = language



from plone.app.multilingual.interfaces import IPloneAppMultilingualInstalled
from plone.app.multilingual.testing import PAM_FUNCTIONAL_TESTING
from plone.base.interfaces import ILanguage
from plone.dexterity.utils import createContentInContainer
from Products.CMFCore.utils import getToolByName
from zope.interface import alsoProvides

import unittest


class TestSubscribers(unittest.TestCase):
    """There are some events that are fired when an object
    is created, moved or copied.

    plone.multilingual registers some subscribers for each event
    to change the language of the object from the container where
    it has been created, moved or copied
    """

    layer = PAM_FUNCTIONAL_TESTING

    def setUp(self):
        alsoProvides(self.layer["request"], IPloneAppMultilingualInstalled)
        self.portal = self.layer["portal"]

    def test_created_event(self):
        """When an object is created in a folder it takes its language from the
        folder itself
        """
        a_ca = createContentInContainer(
            self.portal["ca"], "Document", title="Test document"
        )
        self.assertEqual(ILanguage(a_ca).get_language(), "ca")

    def test_created_event_on_portal(self):
        """When an object is created on portal it should be language
        independent
        """
        a_ca = createContentInContainer(self.portal, "Document", title="Test document")
        self.assertEqual(ILanguage(a_ca).get_language(), "")

    def test_moved_event(self):
        """When an object is moved from within one Language Root Folder into
        a different Language Root Folder it changes its language to that of the
        folder it is copied into
        """
        a_ca = createContentInContainer(
            self.portal["ca"], "Document", title="Test document"
        )

        id_ = self.portal["ca"].manage_cutObjects(a_ca.id)
        self.portal["en"].manage_pasteObjects(id_)
        a_ca_copied = self.portal["en"][a_ca.id]
        self.assertEqual(ILanguage(a_ca_copied).get_language(), "en")

    def test_copied_event(self):
        """When an object is copied from within one Language Root Folder into
        a different Language Root Folder it changes its language to that of the
        folder it is copied into
        """
        a_ca = createContentInContainer(
            self.portal["ca"], "Document", title="Test document"
        )

        id_ = self.portal["ca"].manage_copyObjects(a_ca.id)
        self.portal["en"].manage_pasteObjects(id_)
        a_ca_copied = self.portal["en"][a_ca.id]
        self.assertEqual(ILanguage(a_ca_copied).get_language(), "en")

    def test_moved_to_assets_folder(self):
        """When an object is moved from within one Language Root Folder into
        the Language Independent Folder (named 'Assets') it becomes language
        independent, and it should be visible from the assets folder accessed
        from within other Language Root Folders
        """
        a_ca = createContentInContainer(
            self.portal["ca"], "Document", title="Test document"
        )

        # Test a paste into a subfolder to be ultra cautious
        ca_assets_subfolder = createContentInContainer(
            self.portal["ca"]["recursos"], "Folder", title="A Folder"
        )

        subfolder_name = ca_assets_subfolder.id

        id_ = self.portal["ca"].manage_cutObjects(a_ca.id)
        ca_assets_subfolder.manage_pasteObjects(id_)

        # Get both assets folders afresh
        ca_assets_subfolder = self.portal["ca"]["recursos"][subfolder_name]
        en_assets_subfolder = self.portal["en"]["assets"][subfolder_name]

        # Check it is in both folder listings
        self.assertTrue(a_ca.id in ca_assets_subfolder)
        self.assertTrue(a_ca.id in en_assets_subfolder)

        # Check it is language independent
        copy_in_en = en_assets_subfolder[a_ca.id]
        self.assertEqual(ILanguage(copy_in_en).get_language(), "")
        copy_in_ca = ca_assets_subfolder[a_ca.id]
        self.assertEqual(ILanguage(copy_in_ca).get_language(), "")

        # Check it is returned in catalog search
        catalog = getToolByName(self.portal, "portal_catalog")

        ca_subfolder_path = "/".join(ca_assets_subfolder.getPhysicalPath())
        ca_folder_contents = [r.id for r in catalog(path=ca_subfolder_path)]
        self.assertTrue(a_ca.id in ca_folder_contents)

        en_subfolder_path = "/".join(en_assets_subfolder.getPhysicalPath())
        en_folder_contents = [r.id for r in catalog(path=en_subfolder_path)]
        self.assertTrue(a_ca.id in en_folder_contents)

    def test_copied_to_assets_folder(self):
        """When an object is copied from within one Language Root Folder into
        the Language Independent Folder (named 'Assets') it becomes language
        independent, and it should be visible from the assets folder accessed
        from within other Language Root Folders
        """
        a_ca = createContentInContainer(
            self.portal["ca"], "Document", title="Test document"
        )

        # Test a paste into a subfolder to be ultra cautious
        ca_assets_subfolder = createContentInContainer(
            self.portal["ca"]["recursos"], "Folder", title="A Folder"
        )

        subfolder_name = ca_assets_subfolder.id
        id_ = self.portal["ca"].manage_copyObjects(a_ca.id)
        ca_assets_subfolder.manage_pasteObjects(id_)

        # Get both assets folders afresh
        ca_assets_subfolder = self.portal["ca"]["recursos"][subfolder_name]
        en_assets_subfolder = self.portal["en"]["assets"][subfolder_name]

        # Check it is in both folder listings
        self.assertTrue(a_ca.id in ca_assets_subfolder)
        self.assertTrue(a_ca.id in en_assets_subfolder)

        # Check it is language independent
        copy_in_en = en_assets_subfolder[a_ca.id]
        self.assertEqual(ILanguage(copy_in_en).get_language(), "")
        copy_in_ca = ca_assets_subfolder[a_ca.id]
        self.assertEqual(ILanguage(copy_in_ca).get_language(), "")

        # Check it is returned in catalog search
        catalog = getToolByName(self.portal, "portal_catalog")

        ca_subfolder_path = "/".join(ca_assets_subfolder.getPhysicalPath())
        ca_folder_contents = [r.id for r in catalog(path=ca_subfolder_path)]
        self.assertTrue(a_ca.id in ca_folder_contents)

        en_subfolder_path = "/".join(en_assets_subfolder.getPhysicalPath())
        en_folder_contents = [r.id for r in catalog(path=en_subfolder_path)]
        self.assertTrue(a_ca.id in en_folder_contents)



from plone.app.multilingual import logger
from plone.base.interfaces import ILanguage
from plone.registry.interfaces import IRegistry
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.utils import _createObjectByType
from time import time
from zope.component import getUtility

import transaction


SHARED_NAME = "shared"  # old shared folder name
OLD_PREFIX = "old_"  # temporary prefix while migrating
PROFILE_ID = "profile-plone.app.multilingual:default"


def reimport_css_registry(context):
    setup = getToolByName(context, "portal_setup")
    setup.runImportStepFromProfile(
        "profile-plone.app.multilingual:default",
        "cssregistry",
        run_dependencies=False,
        purge_old=False,
    )

    # Refresh css
    cssregistry = getToolByName(context, "portal_css")
    cssregistry.cookResources()


def migration_pam_1_to_2(context):
    """Migration plone.app.multilingual 1.x to 2.0 by renaming existing
    language folders and creating new LRF containers where existing
    objects are moved into. Old shared content is moved to portal
    root."""

    s1 = time()
    type_name = "LRF"
    ltool = getToolByName(context, "portal_languages")
    utool = getToolByName(context, "portal_url")
    wtool = getToolByName(context, "portal_workflow")
    portal = utool.getPortalObject()

    logger.info("Starting migration of language folders.")

    for code, name in ltool.listSupportedLanguages():
        if code not in portal:
            continue

        older = portal[code]

        if older.portal_type == type_name:
            logger.info(f"'{code}' is alredy a {type_name}, skipping.")
            continue

        # PHASE 1: rename old language folders
        s2 = time()
        old_id = OLD_PREFIX + older.id
        logger.info(f"{code} - Phase 1: Renaming to '{old_id}' ...")
        portal.manage_renameObject(older.id, old_id)
        logger.info(
            "{} - Phase 1: Renaming to '{}' took {:.2f}s.".format(
                code, old_id, time() - s2
            )
        )
        transaction.savepoint()

        # PHASE 2: move content to new LRF
        s3 = time()
        old = portal[old_id]
        logger.info(f"{code} - Phase 2: Moving objects into new LRF...")

        _createObjectByType(type_name, portal, code)
        new = portal[code]
        new.setTitle(name)
        ILanguage(new).set_language(code)

        state = wtool.getInfoFor(new, "review_state", None)
        available_transitions = [t["id"] for t in wtool.getTransitionsFor(new)]
        if state != "published" and "publish" in available_transitions:
            wtool.doActionFor(new, "publish")
        new.reindexObject()
        transaction.savepoint()

        new.manage_pasteObjects(old.manage_cutObjects(ids=old.objectIds()))

        logger.info(
            "{} - Phase 2: Moving objects to LRF took in {:.2f}s.".format(
                code, time() - s3
            )
        )

        transaction.savepoint()

        # PHASE 3: remove old language folders
        s4 = time()
        portal.manage_delObjects(
            ids=[
                old_id,
            ]
        )
        logger.info(
            "{} - Phase 3: Removing '{}' took {:.2f}s.".format(
                code, old_id, time() - s4
            )
        )

        transaction.savepoint()

    # PHASE 4: Old shared folder
    if SHARED_NAME in portal:
        s5 = time()
        shared = portal[SHARED_NAME]
        logger.info(f"{SHARED_NAME} - Phase 4: Moving content to root...")

        portal.manage_pasteObjects(shared.manage_cutObjects(ids=shared.objectIds()))

        logger.info(
            "{} - Phase 4: Moving objects into root took {:.2f}s.".format(
                SHARED_NAME, time() - s5
            )
        )

        transaction.savepoint()

        s6 = time()
        portal.manage_delObjects(
            ids=[
                SHARED_NAME,
            ]
        )
        logger.info(f"{SHARED_NAME} - Phase 5: Removing it took {time() - s6:.2f}s.")

    logger.info(f"All finished in {time() - s1}.")


def upgrade_to_3(context):
    registry = getUtility(IRegistry)

    # don't re-create if it already exists
    key = (
        "plone.app.multilingual.interfaces.IMultiLanguageExtraOptionsSchema."
        "bypass_languageindependent_field_permission_check"
    )
    if key in registry:
        return

    context.runImportStepFromProfile(
        PROFILE_ID.replace("default", "to_3"),
        "plone.app.registry",
    )


def upgrade_to_4(context):
    context.runImportStepFromProfile(
        PROFILE_ID.replace("default", "to_4"),
        "plone.app.registry",
    )



from AccessControl import getSecurityManager
from Acquisition import aq_parent
from plone.app.multilingual import _
from plone.app.multilingual.browser.interfaces import ITranslateMenu
from plone.app.multilingual.browser.interfaces import ITranslateSubMenuItem
from plone.app.multilingual.browser.utils import is_language_independent
from plone.app.multilingual.browser.vocabularies import translated_languages
from plone.app.multilingual.browser.vocabularies import translated_urls
from plone.app.multilingual.browser.vocabularies import untranslated_languages
from plone.app.multilingual.interfaces import ILanguageRootFolder
from plone.app.multilingual.interfaces import IMultiLanguageExtraOptionsSchema
from plone.app.multilingual.interfaces import IPloneAppMultilingualInstalled
from plone.app.multilingual.interfaces import ITG
from plone.app.multilingual.interfaces import ITranslatable
from plone.app.multilingual.interfaces import ITranslationManager
from plone.app.multilingual.interfaces import LANGUAGE_INDEPENDENT
from plone.app.multilingual.permissions import ManageTranslations
from plone.base.interfaces import ILanguage
from plone.memoize import view
from plone.registry.interfaces import IRegistry
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.defaultpage import is_default_page
from Products.CMFPlone.utils import safe_unicode
from zope.browsermenu.menu import BrowserMenu
from zope.browsermenu.menu import BrowserSubMenuItem
from zope.component import getUtility
from zope.component.hooks import getSite
from zope.interface import implementer


@implementer(ITranslateMenu)
class TranslateMenu(BrowserMenu):
    def getMenuItems(self, context, request):
        """Return menu item entries in a TAL-friendly form."""
        # Settings
        site_url = getSite().absolute_url()
        language_tool = getToolByName(context, "portal_languages")
        show_flags = language_tool.showFlags
        try:
            lang_names = request.locale.displayNames.languages
        except AttributeError:
            lang_names = {}

        registry = getUtility(IRegistry)
        settings = registry.forInterface(
            IMultiLanguageExtraOptionsSchema, prefix="plone"
        )
        if settings.redirect_babel_view:
            translation_view = "babel_edit"
        else:
            translation_view = "edit"

        # Content
        content = context
        content_url = context.absolute_url()
        content_language = ILanguage(content).get_language()
        content_translatable = not (
            content_language == LANGUAGE_INDEPENDENT
            or is_language_independent(content)
            or ILanguageRootFolder.providedBy(content)
        )
        content_translated = translated_languages(content)
        content_untranslated = untranslated_languages(content)
        content_translation_group = ITG(content, "")
        if not ITranslatable.providedBy(content):
            content = None

        # Folder when content is default page
        folder = aq_parent(context)
        if not is_default_page(folder, context):
            folder = None
        if folder and ITranslatable.providedBy(folder):  # noqa
            folder_url = folder.absolute_url()
            folder_language = ILanguage(folder).get_language()
            folder_translatable = not (
                folder_language == LANGUAGE_INDEPENDENT
                or is_language_independent(folder)
                or ILanguageRootFolder.providedBy(folder)
            )
            folder_translated = translated_languages(folder)
            folder_untranslated = untranslated_languages(folder)
            folder_translation_group = ITG(folder, "")
        else:
            folder_url = ""
            folder_language = ""
            folder_translatable = False
            folder_translated = []
            folder_untranslated = []
            folder_translation_group = ""

        # Assets folder
        assets_folder_url = None
        assets_folder_title = None
        pc = getToolByName(getSite(), "portal_catalog")
        results = pc.unrestrictedSearchResults(
            portal_type="LIF", Language=ILanguage(context).get_language()
        )
        for brain in results:
            assets_folder_url = brain.getURL() + "/folder_contents"
            assets_folder_title = safe_unicode(brain.Title)
            break

        # Menu items
        results = []
        results_folder = []
        results_content = []

        if folder_translatable:
            # Folder translation view
            lang_name = lang_names.get(folder_language, folder_language)
            results_folder.append(
                {
                    "title": _(
                        "edit_translation",
                        default="Edit ${lang_name}",
                        mapping={"lang_name": lang_name},
                    ),
                    "description": _(
                        "description_babeledit_menu",
                        default="Edit {lang_name} with the two-column translation view",  # noqa
                        mapping={"lang_name": lang_name},
                    ),
                    "action": folder_url + "/" + translation_view,
                    "selected": False,
                    "icon": None,
                    "extra": {
                        "id": "_edit_folder_babel_edit",
                        "separator": None,
                        "class": "",
                    },
                    "submenu": None,
                }
            )

        if (
            folder and folder_untranslated and False
        ):  # disabled in favor of cut & paste  # noqa
            # Set folder language
            results_folder.append(
                {
                    "title": _("title_set_language", default="Change content language"),
                    "description": _(
                        "description_set_language",
                        default="Move the translation under another language folder",  # noqa
                    ),
                    "action": folder_url + "/update_language",
                    "selected": False,
                    "icon": None,
                    "extra": {
                        "id": "_set_folder_language",
                        "separator": None,
                        "class": "",
                    },
                    "submenu": None,
                }
            )

        if folder_translatable:
            for idx, lang in enumerate(folder_untranslated):
                lang_name = lang_names.get(lang.value, lang.title)
                lang_id = lang.value
                icon = (
                    show_flags and language_tool.getFlagForLanguageCode(lang_id) or None
                )  # noqa
                results_folder.append(
                    {
                        "title": _(
                            "create_translation",
                            default="Create ${lang_name}",
                            mapping={"lang_name": lang_name},
                        ),
                        "description": _(
                            "description_translate_into",
                            default="Translate into ${lang_name}",
                            mapping={"lang_name": lang_name},
                        ),
                        "action": "%s/@@create_translation?language=%s"
                        % (folder_url, lang_id),
                        "selected": False,
                        "icon": icon,
                        "width": "14",
                        "height": "11",
                        "extra": {
                            "id": "translate_folder_into_%s" % lang_id,
                            "separator": None,
                            "class": "contentmenuflags",
                        },
                        "submenu": None,
                    }
                )
            urls = translated_urls(folder)
            for lang in folder_translated:
                if lang.value not in urls.by_token:
                    # omit if translation is not permitted to access.
                    continue
                lang_name = lang_names.get(lang.value, lang.title)
                lang_id = lang.value
                icon = (
                    show_flags and language_tool.getFlagForLanguageCode(lang_id) or None
                )  # noqa
                results_folder.append(
                    {
                        "title": _(
                            "edit_translation",
                            default="Edit ${lang_name}",
                            mapping={"lang_name": lang_name},
                        ),
                        "description": _(
                            "description_babeledit_menu",
                            default="Edit {lang_name} with the two-column translation view",  # noqa
                            mapping={"lang_name": lang_name},
                        ),
                        "action": (
                            urls.getTerm(lang_id).title + "/" + translation_view
                        ),
                        "selected": False,
                        "icon": icon,
                        "width": "14",
                        "height": "11",
                        "extra": {
                            "id": "babel_edit_%s" % lang_id,
                            "separator": None,
                            "class": "contentmenuflags",
                        },
                        "submenu": None,
                    }
                )
            # Manage folder translations
            results_folder.append(
                {
                    "title": _(
                        "title_modify_translations", default="Manage translations"
                    ),
                    "description": _(
                        "description_modify_translations",
                        default="Add or delete translations",
                    ),
                    "action": folder_url + "/modify_translations",
                    "selected": False,
                    "icon": None,
                    "extra": {
                        "id": "_modify_folder_translations",
                        "separator": None,
                        "class": "",
                    },
                    "submenu": None,
                }
            )
            # Universal link
            if folder_translation_group:
                results_folder.append(
                    {
                        "title": _("universal_link", default="Universal link"),
                        "description": _(
                            "description_universal_link",
                            default="Universal link to the content in user's preferred language",  # noqa
                        ),
                        "action": "%s/@@multilingual-universal-link/%s"
                        % (site_url, folder_translation_group),
                        "selected": False,
                        "icon": None,
                        "extra": {
                            "id": "_universal_folder_link",
                            "separator": None,
                            "class": "",
                        },
                        "submenu": None,
                    }
                )

        if results_folder:
            # Folder translation separator
            results.append(
                {
                    "title": _(
                        "title_translate_current_folder", default="Folder translation"
                    ),
                    "description": "",
                    "action": None,
                    "selected": False,
                    "icon": None,
                    "extra": {
                        "id": "translateFolderHeader",
                        "separator": "actionSeparator",
                        "class": "plone-toolbar-submenu-header",
                    },
                    "submenu": None,
                }
            )
        results.extend(results_folder)

        lang_name = lang_names.get(content_language, content_language)

        # Content language
        if content_untranslated and False:  # disabled in favor of cut & paste
            results_content.append(
                {
                    "title": _("title_set_language", default="Change content language"),
                    "description": _(
                        "description_set_language",
                        default="Move the translation under another language folder",  # noqa
                    ),
                    "action": content_url + "/update_language",
                    "selected": False,
                    "icon": None,
                    "extra": {"id": "_set_language", "separator": None, "class": ""},
                    "submenu": None,
                }
            )

        if content_translatable:
            # Content translation view
            results_content.append(
                {
                    "title": _(
                        "edit_translation",
                        default="Edit ${lang_name}",
                        mapping={"lang_name": lang_name},
                    ),
                    "description": _(
                        "description_babeledit_menu",
                        default="Edit {lang_name} with the two-column translation view",  # noqa
                        mapping={"lang_name": lang_name},
                    ),
                    "action": content_url + "/" + translation_view,
                    "selected": False,
                    "icon": None,
                    "extra": {
                        "id": "_edit_babel_edit",
                        "separator": None,
                        "class": "",
                    },
                    "submenu": None,
                }
            )

        if content_translatable:
            for idx, lang in enumerate(content_untranslated):
                lang_name = lang_names.get(lang.value, lang.title)
                lang_id = lang.value
                icon = (
                    show_flags and language_tool.getFlagForLanguageCode(lang_id) or None
                )  # noqa
                results_content.append(
                    {
                        "title": _(
                            "create_translation",
                            default="Create ${lang_name}",
                            mapping={"lang_name": lang_name},
                        ),
                        "description": _(
                            "description_translate_into",
                            default="Translate into ${lang_name}",
                            mapping={"lang_name": lang_name},
                        ),
                        "action": "%s/@@create_translation?language=%s"
                        % (content_url, lang_id),
                        "selected": False,
                        "icon": icon,
                        "width": "14",
                        "height": "11",
                        "extra": {
                            "id": "translate_into_%s" % lang_id,
                            "separator": None,
                            "class": "contentmenuflags",
                        },
                        "submenu": None,
                    }
                )
            urls = translated_urls(content)
            for lang in content_translated:
                if lang.value not in urls.by_token:
                    # omit if translation is not permitted to access.
                    continue
                lang_name = lang_names.get(lang.value, lang.title)
                lang_id = lang.value
                icon = (
                    show_flags and language_tool.getFlagForLanguageCode(lang_id) or None
                )  # noqa
                results_content.append(
                    {
                        "title": _(
                            "edit_translation",
                            default="Edit ${lang_name}",
                            mapping={"lang_name": lang_name},
                        ),
                        "description": _(
                            "description_babeledit_menu",
                            default="Edit {lang_name} with the two-column translation view",  # noqa
                            mapping={"lang_name": lang_name},
                        ),
                        "action": (
                            urls.getTerm(lang_id).title + "/" + translation_view
                        ),
                        "selected": False,
                        "icon": icon,
                        "width": "14",
                        "height": "11",
                        "extra": {
                            "id": "babel_edit_%s" % lang_id,
                            "separator": None,
                            "class": "contentmenuflags",
                        },
                        "submenu": None,
                    }
                )
            # Manage content translations
            results_content.append(
                {
                    "title": _(
                        "title_modify_translations", default="Manage translations"
                    ),
                    "description": _(
                        "description_modify_translations",
                        default="Add or delete translations",
                    ),
                    "action": content_url + "/modify_translations",
                    "selected": False,
                    "icon": None,
                    "extra": {
                        "id": "_modify_translations",
                        "separator": None,
                        "class": "",
                    },
                    "submenu": None,
                }
            )
            # Universal link
            if content_translation_group:
                results_content.append(
                    {
                        "title": _("universal_link", default="Universal link"),
                        "description": _(
                            "description_universal_link",
                            default="Universal link to the content in user's preferred language",  # noqa
                        ),
                        "action": "%s/@@multilingual-universal-link/%s"
                        % (site_url, content_translation_group),
                        "selected": False,
                        "icon": None,
                        "extra": {
                            "id": "_universal_link",
                            "separator": None,
                            "class": "",
                        },
                        "submenu": None,
                    }
                )

        if results_folder and results_content:
            # Item translations separator
            results.append(
                {
                    "title": _(
                        "title_translate_current_item", default="Item translation"
                    ),
                    "description": "",
                    "action": None,
                    "selected": False,
                    "icon": None,
                    "extra": {
                        "id": "translateItemHeader",
                        "separator": "actionSeparator",
                        "class": "",
                    },
                    "submenu": None,
                }
            )
        results.extend(results_content)

        # Language independent assets folder
        if assets_folder_url:
            results.append(
                {
                    "title": _(
                        "shared_folder",
                        default="Open ${title} folder",
                        mapping={"title": assets_folder_title},
                    ),
                    "description": _(
                        "description_shared_folder",
                        default="Open the language independent assets folder",
                    ),
                    "action": assets_folder_url,
                    "selected": False,
                    "icon": None,
                    "extra": {
                        "id": "_shared_folder",
                        "separator": results and "actionSeparator" or None,
                        "class": "",
                    },
                    "submenu": None,
                }
            )

        # User preferred language root folder
        if not folder_translatable and not content_translatable:
            results.append(
                {
                    "title": _("language_folder", default="Return to language folder"),
                    "description": _(
                        "description_language_folder",
                        default="Go to the user's browser preferred language "
                        "related folder",
                    ),
                    "action": site_url
                    + "/"
                    + language_tool.getPreferredLanguage(),  # noqa
                    "selected": False,
                    "icon": None,
                    "extra": {
                        "id": "_language_folder",
                        "separator": (
                            (results and not assets_folder_url)
                            and "actionSeparator"
                            or None
                        ),
                        "class": "",
                    },
                    "submenu": None,
                }
            )

        return results


@implementer(ITranslateSubMenuItem)
class TranslateSubMenuItem(BrowserSubMenuItem):
    title = _("label_translate_menu", default="Translate")
    description = _(
        "title_translate_menu", default="Manage translations for your content."
    )
    submenuId = "plone_contentmenu_multilingual"
    order = 5
    extra = {"id": "plone-contentmenu-multilingual"}

    @property
    def action(self):
        return self.context.absolute_url() + "/add_translations"

    @property
    def extra(self):
        return {
            "id": "plone-contentmenu-multilingual",
            "li_class": "plonetoolbar-multilingual",
        }

    @view.memoize
    def available(self):
        # Is PAM installed?
        if not IPloneAppMultilingualInstalled.providedBy(self.request):
            return False

        # Do we have portal_languages?
        lt = getToolByName(self.context, "portal_languages", None)
        if lt is None:
            return False

        # Do we have multiple languages?
        supported = lt.listSupportedLanguages()
        if len(supported) < 2:
            return False

        # And now check permissions
        sm = getSecurityManager()
        if not sm.checkPermission(ManageTranslations, self.context):
            return False

        return True

    def selected(self):
        return False



from plone.app.multilingual import _
from plone.app.multilingual.browser.interfaces import IConnectTranslation
from plone.app.multilingual.interfaces import ITranslationManager
from plone.autoform.form import AutoExtensibleForm
from plone.autoform.interfaces import IFormFieldProvider
from plone.base.interfaces import ILanguage
from plone.registry.interfaces import IRegistry
from plone.uuid.interfaces import IUUID
from Products.CMFCore.utils import getToolByName
from Products.Five.browser import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from Products.statusmessages.interfaces import IStatusMessage
from z3c.form import button
from z3c.form.form import Form
from zope.component import getUtility

import logging


logger = logging.getLogger(__name__)


class ModifyTranslationsForm(BrowserView):
    def available_languages(self):
        registry = getUtility(IRegistry)
        return registry["plone.available_languages"]

    def get_translation(self, language):
        manager = ITranslationManager(self.context)
        return manager.get_translation(language)


class ConnectTranslation(AutoExtensibleForm, Form):
    schema = IFormFieldProvider(IConnectTranslation)
    ignoreContext = True
    label = _("label_connect_translation", default="Connect translation")
    description = _(
        "long_description_connect_translation",
        default="This form allows you to connect a currently existing "
        "translations of the current object.",
    )

    @button.buttonAndHandler(_("connect_translation", default="Connect translation"))
    def handle_add(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        content = data["content"]
        language = data["language"]
        ILanguage(content).set_language(language)
        itm = ITranslationManager(self.context)
        # the 'register_translation'-method takes content OR
        # UUID as second parameter. We need to use the UUID
        # here because otherwise the catalog can't be acquired
        # and the translation index is not updated
        itm.register_translation(language, IUUID(content))
        return self.request.response.redirect(
            self.context.absolute_url() + "/modify_translations"
        )


class DisconnectTranslation(BrowserView):
    tpl = ViewPageTemplateFile("templates/disconnect_translation.pt")

    def __call__(self):
        if self.request.form.get("submitted"):
            language = self.request.form["language"]
            catalog = getToolByName(self.context, "portal_catalog")
            context = catalog.unrestrictedSearchResults(
                UID=self.request.form["came_from"]
            )
            if context:
                context = context[0].getObject()
            if language and context:
                manager = ITranslationManager(context)
                try:
                    manager.remove_translation(language)
                except Exception as e:
                    messages = IStatusMessage(self.request)
                    messages.addStatusMessage(e, type="error")

                return self.request.response.redirect(
                    context.absolute_url() + "/modify_translations"
                )

        return self.tpl()



from plone.app.multilingual.interfaces import IPloneAppMultilingualInstalled
from plone.app.multilingual.interfaces import LANGUAGE_INDEPENDENT
from plone.app.multilingual.testing import PAM_FUNCTIONAL_TESTING
from plone.dexterity.utils import createContentInContainer
from Products.CMFCore.utils import getToolByName
from zope.interface import alsoProvides

import unittest


class TestIndependentLanguageFolder(unittest.TestCase):
    layer = PAM_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        alsoProvides(self.layer["request"], IPloneAppMultilingualInstalled)
        createContentInContainer(
            self.portal["en"]["assets"], "Document", title="Test document"
        )
        # ^ This will be shadowed to all language independent folders

        createContentInContainer(self.portal["ca"], "Document", title="Test document")

        createContentInContainer(self.portal["es"], "Document", title="Test document")

        createContentInContainer(self.portal["en"], "Document", title="Test document")

    def test_query_with_all(self):
        catalog = getToolByName(self.portal, "portal_catalog")

        used_langs = {x.Language for x in catalog.searchResults(self.request)}

        self.assertIn(LANGUAGE_INDEPENDENT, used_langs)
        self.assertIn("es", used_langs)
        self.assertIn("ca", used_langs)
        self.assertIn("en", used_langs)

    def test_query_for_language_independent_content(self):
        catalog = getToolByName(self.portal, "portal_catalog")

        kw = {"Language": LANGUAGE_INDEPENDENT}
        lang_independent = [
            x.Language for x in catalog.searchResults(self.request, **kw)
        ]

        self.assertIn(LANGUAGE_INDEPENDENT, lang_independent)
        self.assertEqual(len(lang_independent), 3)

    def test_query_for_one_language(self):
        catalog = getToolByName(self.portal, "portal_catalog")

        kw = {"Language": "ca"}
        self.assertEqual(
            [x.Language for x in catalog.searchResults(self.request, **kw)],
            ["ca", "ca", "ca"],
        )



from plone.dexterity.browser.view import DefaultView
from z3c.form.interfaces import IEditForm
from zope.interface import implementer


@implementer(IEditForm)
class DexterityBabelView(DefaultView):
    # By claiming that we implement IEditForm, we also get the TaggedValues of
    # that interface. This is important, because this view is displayed side
    # by side with the edit from, and the DefaultView itself has some tagged
    # values to hide fields.
    pass



from plone.app.multilingual.interfaces import ITG
from plone.app.multilingual.interfaces import ITranslatable
from plone.app.multilingual.interfaces import ITranslationManager


def get_translation_group(content):
    """Get the translation group.

    :param content: Content object which is part of the translation group.
    :type content: Content object
    :returns: UID string identifier of the translation group
    :raises:
        ValueError
    """
    tg = ITG(content)
    if tg is None:
        raise ValueError("No translation group found.")
    return tg


def get_translation_manager(content):
    """Get the translation manager.

    :param content: Content for which the tranlation manager is needed.
    :type content: Content object
    :returns: translation manager instance.
    :raises:
        ValueError
    """
    tm = ITranslationManager(content)
    if tm is None:
        raise ValueError("No translation manager available for this content.")
    return tm


def translate(content, target_language="en"):
    """Translate content into target language.

    :param content: Content to be translated.
    :type content: Content object
    :param target_language: Language to be translated to.
    :type target_language: String
    :returns: Content object in new language
    """
    manager = get_translation_manager(content)
    manager.add_translation(target_language)
    return manager.get_translation(target_language)


def is_translatable(content):
    """Checks if content is translatable.

    :param content: Content to be translated.
    :type content: Content object
    :returns: Boolean
    """
    return ITranslatable.providedBy(content)



from gzip import GzipFile
from io import BytesIO
from plone.app.multilingual.interfaces import IPloneAppMultilingualInstalled
from plone.app.multilingual.testing import PAM_FUNCTIONAL_TESTING
from plone.base.interfaces import ISiteSchema
from plone.dexterity.utils import createContentInContainer
from plone.registry.interfaces import IRegistry
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.interface import alsoProvides

import unittest


# This class largely inspired by plone/app/layout/sitemap/tests/test_sitemap.py
class TestSitemap(unittest.TestCase):
    layer = PAM_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        alsoProvides(self.layer["request"], IPloneAppMultilingualInstalled)

        registry = getUtility(IRegistry)
        self.site_settings = registry.forInterface(ISiteSchema, prefix="plone")
        self.site_settings.enable_sitemap = True

        self.sitemap = getMultiAdapter(
            (self.portal, self.portal.REQUEST), name="sitemap.xml.gz"
        )

        createContentInContainer(
            self.portal["en"]["assets"], "Document", title="Test document"
        )
        # ^ This will be shadowed to all language independent folders

        createContentInContainer(self.portal["ca"], "Document", title="Test document")

        createContentInContainer(self.portal["es"], "Document", title="Test document")

        createContentInContainer(self.portal["en"], "Document", title="Test document")

    def uncompress(self, sitemapdata):
        sio = BytesIO(sitemapdata)
        unziped = GzipFile(fileobj=sio)
        xml = unziped.read()
        unziped.close()
        return xml

    def test_portalroot_sitemap(self):
        """
        Requests for the sitemap on portalroot return all languages
        """

        xml = self.uncompress(self.sitemap())
        self.assertIn(b"<loc>http://nohost/plone/ca/test-document</loc>", xml)
        self.assertIn(b"<loc>http://nohost/plone/en/test-document</loc>", xml)
        self.assertIn(b"<loc>http://nohost/plone/es/test-document</loc>", xml)

        self.assertIn(b"<loc>http://nohost/plone/ca/recursos/test-document</loc>", xml)
        self.assertIn(b"<loc>http://nohost/plone/en/assets/test-document</loc>", xml)
        self.assertIn(b"<loc>http://nohost/plone/es/recursos/test-document</loc>", xml)

    def test_navroot_sitemap(self):
        """
        Sitemap generated from a LanguageRootFolder (an INavigationRoot)
        """
        sitemap = getMultiAdapter(
            (self.portal.es, self.portal.REQUEST), name="sitemap.xml.gz"
        )
        xml = self.uncompress(sitemap())
        self.assertNotIn(
            b"<loc>http://nohost/plone/ca/test-document</loc>", xml
        )  # noqa
        self.assertNotIn(
            b"<loc>http://nohost/plone/en/test-document</loc>", xml
        )  # noqa
        self.assertIn(b"<loc>http://nohost/plone/es/test-document</loc>", xml)

        self.assertNotIn(
            b"<loc>http://nohost/plone/ca/recursos/test-document</loc>", xml
        )
        self.assertNotIn(b"<loc>http://nohost/plone/en/assets/test-document</loc>", xml)
        self.assertIn(b"<loc>http://nohost/plone/es/recursos/test-document</loc>", xml)



from Acquisition import aq_inner
from plone.app.multilingual import _
from plone.app.multilingual.interfaces import IMultiLanguageExtraOptionsSchema
from plone.app.multilingual.interfaces import ITranslationManager
from plone.base.interfaces import ILanguage
from plone.registry.interfaces import IRegistry
from plone.uuid.interfaces import IUUID
from Products.Five import BrowserView
from zope.component import getUtility

import json
import urllib


def google_translate(question, key, lang_target, lang_source):
    length = len(question)
    translated = ""
    url = "https://www.googleapis.com/language/translate/v2"
    temp_question = question
    while length > 400:
        temp_question = question[:399]
        index = temp_question.rfind(" ")
        temp_question = temp_question[:index]
        question = question[index:]
        length = len(question)
        data = {
            "key": key,
            "target": lang_target,
            "source": lang_source,
            "q": temp_question,
        }
        params = urllib.parse.urlencode(data)

        retorn = urllib.request.urlopen(url + "?" + params)
        translated += json.loads(retorn.read())["data"]["translations"][0][
            "translatedText"
        ]

    data = {
        "key": key,
        "target": lang_target,
        "source": lang_source,
        "q": temp_question,
    }
    params = urllib.parse.urlencode(data)

    retorn = urllib.request.urlopen(url + "?" + params)
    translated += json.loads(retorn.read())["data"]["translations"][0]["translatedText"]
    return json.dumps({"data": translated})


class gtranslation_service_dexterity(BrowserView):
    def __call__(self):
        if self.request.method != "POST" and not (
            "field" in self.request.form.keys()
            and "lang_source" in self.request.form.keys()
        ):
            return _("Need a field")
        else:
            manager = ITranslationManager(self.context)
            registry = getUtility(IRegistry)
            settings = registry.forInterface(
                IMultiLanguageExtraOptionsSchema, prefix="plone"
            )
            lang_target = ILanguage(self.context).get_language()
            lang_source = self.request.form["lang_source"]
            orig_object = manager.get_translation(lang_source)
            field = self.request.form["field"].split(".")[-1]
            if hasattr(orig_object, field):
                question = getattr(orig_object, field, "")
                if hasattr(question, "raw"):
                    question = question.raw
            else:
                return _("Invalid field")
            return google_translate(
                question, settings.google_translation_key, lang_target, lang_source
            )


class TranslationForm(BrowserView):
    """Translation Form"""

    def __call__(self):
        language = self.request.get("language", None)
        if language:
            context = aq_inner(self.context)
            translation_manager = ITranslationManager(context)
            new_parent = translation_manager.add_translation_delegated(language)  # noqa
            baseUrl = new_parent.absolute_url()
            url = f"{baseUrl}/++addtranslation++{IUUID(context)}"
            return self.request.response.redirect(url)



from plone.app.multilingual.dx.interfaces import ILanguageIndependentField
from plone.app.multilingual.interfaces import ILanguageIndependentFieldsManager
from plone.app.multilingual.interfaces import ITranslationCloner
from plone.app.multilingual.interfaces import ITranslationManager
from plone.base.interfaces import ILanguage
from plone.dexterity.utils import iterSchemata
from Products.CMFPlone.utils import safe_unicode
from z3c.relationfield import RelationValue
from z3c.relationfield.interfaces import IRelationList
from z3c.relationfield.interfaces import IRelationValue
from zope.component import getUtility
from zope.component import queryAdapter
from zope.interface import implementer
from zope.intid.interfaces import IIntIds


_marker = object()


@implementer(ITranslationCloner)
class Cloner:
    def __init__(self, context):
        self.context = context

    def __call__(self, obj):
        ILanguageIndependentFieldsManager(self.context).copy_fields(obj)


@implementer(ILanguageIndependentFieldsManager)
class LanguageIndependentFieldsManager:
    def __init__(self, context):
        self.context = context

    def has_independent_fields(self):
        for schema in iterSchemata(self.context):
            for field_name in schema:
                if ILanguageIndependentField.providedBy(schema[field_name]):
                    return True
        return False

    def copy_relation(self, relation_value, target_language):
        if not relation_value or relation_value.isBroken():
            return

        obj = relation_value.to_object
        intids = getUtility(IIntIds)
        translation = ITranslationManager(obj).get_translation(target_language)
        if translation:
            return RelationValue(intids.getId(translation))
        return RelationValue(intids.getId(obj))

    def copy_fields(self, translation):
        changed = False

        target_language = queryAdapter(translation, ILanguage).get_language()

        for schema in iterSchemata(self.context):
            context_adapter = None
            translation_adapter = None
            for field_name in schema:
                if ILanguageIndependentField.providedBy(schema[field_name]):
                    if context_adapter is None:
                        context_adapter = schema(self.context)
                    value = getattr(context_adapter, field_name, _marker)
                    field_changed = None
                    if value == _marker:
                        continue
                    elif IRelationValue.providedBy(value):
                        field_changed = True
                        value = self.copy_relation(value, target_language)
                    elif IRelationList.providedBy(schema[field_name]):
                        field_changed = True
                        if not value:
                            value = []
                        else:
                            new_value = []
                            for relation in value:
                                copied_relation = self.copy_relation(
                                    relation, target_language
                                )
                                if copied_relation:
                                    new_value.append(copied_relation)
                            value = new_value

                    if translation_adapter is None:
                        translation_adapter = schema(translation)

                    # We only want to store a new value if it has changed.
                    # In general we can compare equality of the new value to the one on the translation.
                    # But RelationValue.__eq__ is broken if the relation doesn't have a from_object,
                    # so for now we force field_changed to True for relations above.
                    if field_changed is None:
                        translation_value = getattr(
                            translation_adapter, field_name, _marker
                        )
                        field_changed = value != translation_value
                    if field_changed:
                        changed = True
                        setattr(translation_adapter, field_name, safe_unicode(value))

        # If at least one field has been copied over to the translation
        # we need to inform subscriber to trigger an ObjectModifiedEvent
        # on that translation.
        return changed



from OFS.event import ObjectWillBeRemovedEvent
from plone.app.multilingual import api
from plone.app.multilingual.interfaces import ATTRIBUTE_NAME
from plone.app.multilingual.interfaces import IPloneAppMultilingualInstalled
from plone.app.multilingual.interfaces import ITranslationIdChooser
from plone.app.multilingual.interfaces import ITranslationLocator
from plone.app.multilingual.interfaces import ITranslationManager
from plone.app.multilingual.testing import PAM_FUNCTIONAL_TESTING
from plone.app.testing import logout
from plone.base.interfaces import ILanguage
from plone.dexterity.utils import createContentInContainer
from Products.CMFCore.utils import getToolByName
from zope.event import notify
from zope.interface import alsoProvides

import unittest


class TestAPI(unittest.TestCase):
    layer = PAM_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        alsoProvides(self.layer["request"], IPloneAppMultilingualInstalled)

    def test_get_translation_manager(self):
        a_ca = createContentInContainer(
            self.portal["ca"], "Document", title="Test document"
        )
        tm = api.get_translation_manager(a_ca)
        self.assertTrue(ITranslationManager.providedBy(tm))

    def test_translate(self):
        # Create
        a_ca = createContentInContainer(
            self.portal["ca"], "Document", title="Test document"
        )

        self.assertEqual(
            api.get_translation_manager(a_ca).get_translations(), {"ca": a_ca}
        )

        # Translate
        a_en = api.translate(a_ca, "en")

        # check
        self.assertEqual(
            api.get_translation_manager(a_ca).get_translations(),
            {"ca": a_ca, "en": a_en},
        )

    def test_get_translation_group(self):
        # Create
        a_ca = createContentInContainer(
            self.portal["ca"], "Document", title="Test document"
        )
        a_en = api.translate(a_ca, "en")

        # get groups
        tg_ca = api.get_translation_group(a_en)
        tg_en = api.get_translation_group(a_en)

        # check
        self.assertTrue(tg_ca is not None)
        self.assertTrue(tg_en is not None)
        self.assertEqual(tg_en, tg_ca)

    def test_translateable(self):
        # Create
        a_ca = createContentInContainer(
            self.portal["ca"], "Document", title="Test document"
        )
        # check
        self.assertTrue(api.is_translatable(a_ca))
        self.assertFalse(api.is_translatable(object()))


class TestBasicAPI(unittest.TestCase):
    layer = PAM_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        alsoProvides(self.layer["request"], IPloneAppMultilingualInstalled)
        # Create
        self.a_ca = createContentInContainer(
            self.portal["ca"], "Document", title="Test document"
        )

    def test_supported_languages(self):
        language_tool = getToolByName(self.portal, "portal_languages")
        self.assertEqual(language_tool.getSupportedLanguages(), ["en", "ca", "es"])

    def test_get_translations(self):
        translations = ITranslationManager(self.a_ca).get_translations()
        self.assertEqual(translations, {"ca": self.a_ca})

    def test_get_translation(self):
        a_ca = ITranslationManager(self.a_ca).get_translation("ca")
        self.assertEqual(a_ca, self.a_ca)

    def test_get_translated_languages(self):
        translated_languages = ITranslationManager(self.a_ca).get_translated_languages()
        self.assertEqual(translated_languages, ["ca"])

    def test_has_translation(self):
        translation_manager = ITranslationManager(self.a_ca)
        self.assertTrue(translation_manager.has_translation("ca"))
        self.assertFalse(translation_manager.has_translation("es"))

    def test_get_restricted_translation(self):
        restricted_translations = ITranslationManager(
            self.a_ca
        ).get_restricted_translations()
        self.assertEqual(restricted_translations, {"ca": self.a_ca})

    def test_get_restricted_translation_for_anonymous(self):
        logout()
        restricted_translations = ITranslationManager(
            self.a_ca
        ).get_restricted_translations()
        self.assertEqual(restricted_translations, {})

    def test_add_translation(self):
        # Check that document does not exists
        self.assertNotIn("test-document", self.portal["es"].objectIds())

        # Create es translation
        ITranslationManager(self.a_ca).add_translation("es")

        # Check if it exists
        self.assertIn("test-document", self.portal["es"].objectIds())

        # Check language
        language = ILanguage(self.portal["es"]["test-document"]).get_language()
        self.assertEqual(language, "es")

    def test_add_translation_delegated(self):
        # Create es translation
        portal_es = ITranslationManager(self.a_ca).add_translation_delegated("es")
        self.assertEqual(portal_es, self.portal["es"])

    def test_register_translation(self):
        a_es = createContentInContainer(
            self.portal["es"], "Document", title="Test document"
        )

        ITranslationManager(self.a_ca).register_translation("es", a_es)

        translations = ITranslationManager(self.a_ca).get_translations()
        self.assertEqual(translations, {"ca": self.a_ca, "es": a_es})


class TestLanguageRootFolderAPI(unittest.TestCase):
    layer = PAM_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        alsoProvides(self.layer["request"], IPloneAppMultilingualInstalled)

    def test_initial_language_set(self):
        # Create
        a_ca = createContentInContainer(
            self.portal["ca"], "Document", title="Test document"
        )

        # Check that the content has language
        self.assertEqual(ILanguage(a_ca).get_language(), "ca")

    def test_add_translation(self):
        # Create
        a_ca = createContentInContainer(
            self.portal["ca"], "Document", title="Test document"
        )

        # Translate
        ITranslationManager(a_ca).add_translation("es")
        a_es = ITranslationManager(a_ca).get_translation("es")

        # Check that translation is registered
        self.assertEqual(
            ITranslationManager(a_ca).get_translations(), {"ca": a_ca, "es": a_es}
        )

        # Check that it is in the correct folder
        self.assertTrue(a_es.id in self.portal["es"])

        # Check that it cannot be translated again
        translation_manager = ITranslationManager(a_ca)
        self.assertRaises(KeyError, translation_manager.add_translation, "es")

        # Check that language must be valid
        self.assertRaises(KeyError, translation_manager.add_translation, None)

    def test_add_translation_delegated(self):
        a_ca = createContentInContainer(
            self.portal["ca"], "Document", title="Test document"
        )

        translation_manager = ITranslationManager(a_ca)
        portal_es = translation_manager.add_translation_delegated("es")

        self.assertEqual(portal_es, self.portal["es"])

    def test_create_destroy_link_translations(self):
        a_ca = createContentInContainer(
            self.portal["ca"], "Document", title="Test document"
        )

        ITranslationManager(a_ca).add_translation("es")
        a_es = ITranslationManager(a_ca).get_translation("es")

        # Create duplicate content
        another_es = createContentInContainer(
            self.portal["es"], "Document", title="Test another"
        )

        ITranslationManager(another_es).add_translation("ca")
        another_ca = ITranslationManager(another_es).get_translation("ca")

        # Delete original content
        notify(ObjectWillBeRemovedEvent(a_es))
        self.portal.es.manage_delObjects(a_es.id)

        notify(ObjectWillBeRemovedEvent(another_ca))
        self.portal.ca.manage_delObjects(another_ca.id)

        # Check canonical values are still different
        id_a_ca = ITranslationManager(a_ca).query_canonical()
        id_another_es = ITranslationManager(another_es).query_canonical()

        self.assertNotEqual(id_another_es, id_a_ca)
        self.assertTrue(isinstance(id_another_es, str))
        self.assertTrue(isinstance(id_a_ca, str))

        # Make documents translations of each other
        ITranslationManager(a_ca).register_translation("es", another_es)

        # Check that canonical values are now the same
        id_a_ca = ITranslationManager(a_ca).query_canonical()
        id_another_es = ITranslationManager(another_es).query_canonical()

        self.assertEqual(id_another_es, id_a_ca)

    def test_create_relink_translations(self):
        """We check the update function here"""
        a_ca = createContentInContainer(
            self.portal["ca"], "Document", title="Test document"
        )

        b_ca = createContentInContainer(
            self.portal["ca"], "Document", title="Test document"
        )

        b_es = createContentInContainer(
            self.portal["es"], "Document", title="Test document"
        )
        ITranslationManager(b_es).add_translation("ca")
        ITranslationManager(a_ca).register_translation("es", b_es)

        self.assertFalse(ITranslationManager(b_ca).has_translation("es"))
        self.assertEqual(ITranslationManager(b_es).get_translation("ca"), a_ca)

    def test_id_chooser(self):
        a_ca = createContentInContainer(
            self.portal["ca"], "Document", title="Test document"
        )

        chooser = ITranslationIdChooser(a_ca)
        self.assertEqual(chooser(self.portal, "es"), "test-document")

        createContentInContainer(self.portal, "Document", title="Another test")

        b_ca = createContentInContainer(
            self.portal["ca"], "Document", title="Another test"
        )

        chooser = ITranslationIdChooser(b_ca)
        self.assertEqual(chooser(self.portal, "es"), "another-test-es")

    def test_locator(self):
        folder_ca = createContentInContainer(
            self.portal["ca"], "Folder", title="Test folder"
        )
        subfolder_ca = createContentInContainer(
            self.portal["ca"]["test-folder"], "Folder", title="Test folder"
        )

        locator = ITranslationLocator(subfolder_ca)
        self.assertEqual(locator("es"), self.portal["es"])

        ITranslationManager(folder_ca).add_translation("es")
        folder_es = ITranslationManager(folder_ca).get_translation("es")

        child_locator = ITranslationLocator(subfolder_ca)
        self.assertEqual(child_locator("es"), folder_es)

    def test_tg_view(self):
        a_ca = createContentInContainer(
            self.portal["ca"], "Document", title="Test document"
        )
        tg = getattr(a_ca, ATTRIBUTE_NAME)
        self.assertTrue(bool(tg))
        self.assertEqual(a_ca.restrictedTraverse("@@tg")(), tg)



from Acquisition import aq_base
from Acquisition import aq_inner
from Acquisition import aq_parent
from plone.app.multilingual import _
from plone.app.multilingual.interfaces import ITranslationManager
from plone.base.interfaces import ILanguage
from plone.base.interfaces import IPloneSiteRoot
from plone.locking.interfaces import ILockable
from Products.CMFCore.exceptions import ResourceLockedError
from Products.CMFCore.utils import getToolByName
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from zc.relation.interfaces import ICatalog
from zope.component import getUtility
from zope.component.hooks import getSite
from zope.interface import Interface
from zope.interface.interfaces import ComponentLookupError

import logging
import warnings


try:
    from Products.LinguaPlone.interfaces import ITranslatable
except ImportError:
    from plone.app.multilingual.interfaces import ITranslatable

try:
    from plone.dexterity.interfaces import IDexterityContent
except ImportError:

    class IDexterityContent(Interface):
        pass


try:
    from Products.Archetypes.interfaces import IBaseObject
except ImportError:

    class IBaseObject(Interface):
        pass


with warnings.catch_warnings():
    warnings.filterwarnings(
        "ignore",
        message="LanguageRootFolder: LanguageRootFolders should be migrate "
        "to DexterityContainers",
    )
    from plone.app.multilingual.content.lrf import LanguageRootFolder


LP_TRANSLATABLE = "Products.LinguaPlone.interfaces.ITranslatable"
portal_types_blacklist = ["Collage", "FormFolder", "Ploneboard"]
logger = logging.getLogger(__name__)


class LP2PAMView(BrowserView):
    """View for migrating multilingual catalog from LP to PAM."""

    template = ViewPageTemplateFile("templates/migrator-results.pt")
    stepinfo = _("Transfer multilingual catalog information")

    def __call__(self):
        pc = getToolByName(self.context, "portal_catalog")
        pl = getToolByName(self.context, "portal_languages")
        self.results = []
        for language_supported in pl.getSupportedLanguages():
            translated_objects = pc.searchResults(
                object_provides=LP_TRANSLATABLE, Language=language_supported
            )
            for brain in translated_objects:
                obj = brain.getObject()
                if obj.isCanonical():
                    translations = obj.getTranslations(include_canonical=False)
                    manager = ITranslationManager(obj)
                    if translations:
                        for language in translations.keys():
                            try:
                                manager.register_translation(
                                    language, translations[language][0]
                                )
                            except KeyError:
                                logger.info(
                                    "%s already translated to %s: %s"
                                    % (
                                        obj.id,
                                        language,
                                        str(manager.get_translations()),
                                    )
                                )

                        self.results.append(str(manager.get_translations()))

        logger.info("Finished with transferring catalog information")
        return self.template()


class LP2PAMAfterView(BrowserView):
    template = ViewPageTemplateFile("templates/cleanup_results.pt")
    stepinfo = _("After migration relation cleanup")

    def reset_relation_catalog(self):
        """Sometimes there are dependencies to the ITranslatable interface
        hidden in the relation catalog. This reset gets rid of them. (Assuming
        that Products.LinguaPlone is already uninstalled).
        """
        try:
            catalog = getUtility(ICatalog)
        except ComponentLookupError:
            return 0, ["A zc.relation catalog is not present."]
        relations = catalog.findRelations()
        catalog.clear()
        total = 0
        bad = []
        for relation in relations:
            total += 1
            try:
                catalog.index(relation)
            except KeyError:
                # If you read this because you wonder why you have many
                # missing relations, please inform do3cc
                logger.warning(
                    "A relation could not be recreated. You have " "lost relations"
                )
                bad.append(str(relation.__dict__))
        return total, bad

    def __call__(self):
        total, bad = self.reset_relation_catalog()
        self.total = total
        self.bad = bad
        return self.template()


class moveContentToProperRLF(BrowserView):
    """This browser view moves the site's content to its corresponding root
    language folder and previously made a search for misplaced content through
    the site's content tree and moves them to its nearest translated parent.
    """

    template = ViewPageTemplateFile("templates/relocate-results.pt")
    stepinfo = _("Relocate content to the proper root language folder")
    blacklist = list()

    def findContent(self, content, depth):
        # only handle portal content
        if not IDexterityContent.providedBy(content) and not IBaseObject.providedBy(
            content
        ):
            logger.warning(
                "SKIP non-portal content %s (%s)"
                % (content.absolute_url(), content.meta_type)
            )
            return
        if (
            hasattr(aq_base(content), "objectIds")
            and aq_base(content).portal_type not in self.blacklist
        ):
            for id in content.objectIds():
                self.findContent(getattr(content, id), depth + 1)
        while len(self.content_tree) < depth + 1:
            self.content_tree.append([])
        if ITranslatable.providedBy(content):
            # The content parent has the same language?
            if (
                not IPloneSiteRoot.providedBy(aq_parent(content))
                and aq_parent(content).Language() != content.Language()
            ):
                self.content_tree[depth].append(content)

    def searchNearestTranslatedParent(self, content):
        parent = aq_parent(content)
        while parent.Language() != content.Language() and not IPloneSiteRoot.providedBy(
            parent
        ):
            parent = aq_parent(parent)
        return parent

    def __call__(self):
        """Note: Steps names don't correspond with the control panel ones"""
        blacklist = self.request.form.get("blacklist", "").split()
        self.blacklist = [x.strip() for x in blacklist if x.strip() != ""]
        self.results = self.step1andstep2()
        self.results += self.step3()
        return self.template()

    def step1andstep2(self):
        """Explore the site's content searching for misplaced content and move
        it to its nearest translated parent.
        """
        portal = getSite()

        output = []
        # Step 1 - Audit the content tree and make a list with the candidates
        # to be moved to the right RLF. Once we get a candidate, decide if it
        # should be moved to its nearest parent with the same language. Trying
        # to avoid the catalog in order to avoid problems with big sites and
        # bad or corrupted catalogs.
        self.content_tree = []
        self.findContent(portal, 0)
        logger.info("Step 1: Eligible content: %s" % self.content_tree)

        # We now have a list of lists that maps each eligible content with its
        # depth in the content tree

        # Step 2 - Move the eligible content to its nearest translated parent
        # from the most deepest located content to the outer ones
        self.content_tree.reverse()

        for depth in self.content_tree:
            if depth != []:
                for content in depth:
                    parent = aq_parent(content)
                    target_folder = self.searchNearestTranslatedParent(content)
                    # Test if the id already exist previously

                    try:
                        cutted = parent.manage_cutObjects(content.getId())
                    except ResourceLockedError:
                        lockable = ILockable(content)
                        lockable.unlock()
                        cutted = parent.manage_cutObjects(content.getId())
                    try:
                        target_folder.manage_pasteObjects(cutted)
                        info_str = "Step 2: Moved object {} to folder {}".format(
                            "/".join(content.getPhysicalPath()),
                            "/".join(target_folder.getPhysicalPath()),
                        )
                        log = logger.info
                    except Exception as err:
                        info_str = (
                            "ERROR. Step 2: not possible to move "
                            "object %s to folder %s. Error: %s"
                            % (
                                "/".join(content.getPhysicalPath()),
                                "/".join(target_folder.getPhysicalPath()),
                                err,
                            )
                        )
                        log = logger.error
                    log(info_str)
                    output.append(info_str)

        logger.info("Finished step 2")
        return output

    def step3(self):
        """Move the existing site content to its correspondent RLF."""
        portal = getSite()
        pc = getToolByName(portal, "portal_catalog")
        pl = getToolByName(portal, "portal_languages")

        supported_langs = pl.getSupportedLanguages()

        output = []
        # Step 3: Move all the remaining content to its correspondent RLFs
        for lang in supported_langs:
            RLF_id = "%s" % lang
            folder = getattr(portal, RLF_id, None)
            if not folder:
                raise AttributeError(
                    "One of the root language folder are \
                                      missing. Check the site's language \
                                      setup."
                )

            path = "/".join(portal.getPhysicalPath())
            objects = pc.searchResults(
                path={"query": path, "depth": 1},
                sort_on="getObjPositionInParent",
                Language=lang,
            )

            for brain in objects:
                if brain.id != lang:
                    old_path = brain.getPath()

                    try:
                        cutted = self.context.manage_cutObjects(brain.id)
                    except ResourceLockedError:
                        content = brain.getObject()
                        lockable = ILockable(content)
                        lockable.unlock()
                        cutted = self.context.manage_cutObjects(brain.id)
                    try:
                        folder.manage_pasteObjects(cutted)
                        info_str = "Moved object %s to language root folder " "%s" % (
                            old_path,
                            lang,
                        )
                        log = logger.info
                    except Exception as err:
                        info_str = (
                            "ERROR. Step 3: not possible to move "
                            "object %s to root language folder %s. Error: %s"
                            % (old_path, lang, err)
                        )
                        log = logger.error
                    log(info_str)
                    output.append(info_str)

        logger.info("Finished step 3")
        return output


class LP2PAMReindexLanguageIndex(BrowserView):
    template = ViewPageTemplateFile("templates/reindex-results.pt")
    stepinfo = "Reindex the LanguageIndex"

    def __call__(self):
        pc = getToolByName(self.context, "portal_catalog")
        index = pc._catalog.getIndex("Language")
        self.items_before = index.numObjects()
        pc.manage_reindexIndex(ids=["Language"])
        self.items_after = index.numObjects()

        return self.template()


class MigrateFolderToLRFView(BrowserView):
    def __call__(self):
        plone_utils = getToolByName(self.context, "plone_utils")

        if self.context.__class__ == LanguageRootFolder:
            self.request.response.redirect(self.context.absolute_url())
            return

        if not IPloneSiteRoot.providedBy(aq_parent(aq_inner(self.context))):
            plone_utils.addPortalMessage(
                _(
                    "folder_to_lrf_not_next_to_root",
                    default="Only folders just below the root " "can be transformed",
                )
            )
            self.request.response.redirect(self.context.absolute_url())
            return

        portal_languages = getToolByName(self.context, "portal_languages")
        available_languages = portal_languages.getAvailableLanguages()
        if self.context.id not in available_languages.keys():
            plone_utils.addPortalMessage(
                _(
                    "folder_to_lrf_id_not_language",
                    default="Folder's id is not a valid language code",
                )
            )
            self.request.response.redirect(self.context.absolute_url())
            return

        # Do the transform
        self.context.__class__ = LanguageRootFolder
        self.context._p_changed = aq_parent(self.context).p_changed = True
        self.context.portal_type = "LRF"

        # Update content language
        portal_catalog = getToolByName(self.context, "portal_catalog")
        search_results = portal_catalog.unrestrictedSearchResults(
            path="/".join(self.context.getPhysicalPath())
        )
        for brain in search_results:
            ob = brain._unrestrictedGetObject()
            language_aware = ILanguage(ob, None)
            if language_aware is not None:
                language_aware.set_language(self.context.id)
                ob.reindexObject(idxs=["Language", "TranslationGroup"])

        plone_utils.addPortalMessage(
            _(
                "folder_to_lrf_success",
                default="Folder has been successfully transformed to "
                "a language root folder",
            )
        )
        self.request.response.redirect(self.context.absolute_url())



from zope.i18nmessageid import MessageFactory

import logging
import pkg_resources


try:
    pkg_resources.get_distribution("Products.LinguaPlone")
except pkg_resources.DistributionNotFound:
    isLPinstalled = False
else:
    isLPinstalled = True

logger = logging.getLogger("plone.app.multilingual")
_ = MessageFactory("plone")



from plone.app.multilingual import api
from plone.app.multilingual.browser.viewlets import AlternateLanguagesViewlet
from plone.app.multilingual.interfaces import IPloneAppMultilingualInstalled
from plone.app.multilingual.interfaces import ITranslationManager
from plone.app.multilingual.testing import PAM_FUNCTIONAL_TESTING
from plone.dexterity.utils import createContentInContainer
from zope.interface import alsoProvides

import unittest


class TestAlternateLanguagesViewlet(unittest.TestCase):
    layer = PAM_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        alsoProvides(self.layer["request"], IPloneAppMultilingualInstalled)

    def test_alternates(self):
        # Create
        a_ca = createContentInContainer(
            self.portal["ca"], "Document", title="Test document"
        )

        # Translate
        a_en = api.translate(a_ca, "en")

        # Get the viewlet
        viewlet = AlternateLanguagesViewlet(a_en, self.request, None, None)
        viewlet.update()

        # Get translations
        translations = ITranslationManager(a_ca).get_translations()

        # Check translations
        self.assertEqual(len(viewlet.alternates), 2)
        for item in viewlet.alternates:
            self.assertIn(item["lang"], translations)
            self.assertEqual(item["url"], translations[item["lang"]].absolute_url())



from OFS.event import ObjectWillBeRemovedEvent
from plone.app.multilingual.browser.utils import is_language_independent
from plone.app.multilingual.browser.utils import multilingualMoveObject
from plone.app.multilingual.interfaces import IPloneAppMultilingualInstalled
from plone.app.multilingual.interfaces import ITranslationManager
from plone.app.multilingual.testing import PAM_FUNCTIONAL_TESTING
from plone.dexterity.utils import createContentInContainer
from plone.uuid.interfaces import IUUID
from zope.event import notify
from zope.interface import alsoProvides

import unittest


class TestLanguageRootFolder(unittest.TestCase):
    layer = PAM_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        alsoProvides(self.layer["request"], IPloneAppMultilingualInstalled)

    def test_shared_content(self):
        # Create shared document
        createContentInContainer(
            self.portal.en.assets, "Document", title="Test document"
        )

        # Check shared document is there
        self.assertEqual(
            self.portal.en.assets["test-document"],
            self.portal.ca.recursos["test-document"],
        )
        self.assertEqual(
            self.portal.en.assets["test-document"],
            self.portal.es.recursos["test-document"],
        )

        # Delete shared document
        notify(ObjectWillBeRemovedEvent(self.portal.en.assets["test-document"]))
        self.portal.en.assets.manage_delObjects("test-document")

        # Check that it is not available in LRFs
        self.assertNotIn("test-document", self.portal.ca.recursos.objectIds())
        self.assertNotIn("test-document", self.portal.es.recursos.objectIds())

    def test_shared_content_indexing(self):
        # Create shared document
        createContentInContainer(
            self.portal.en.assets, "Document", title="Test document"
        )

        # Check that shared document is indexed in all LRFs
        elements = self.portal.portal_catalog.searchResults(id="test-document")
        self.assertEqual(len(elements), 3)

        # Remove shared document
        notify(ObjectWillBeRemovedEvent(self.portal.en.assets["test-document"]))
        self.portal.en.assets.manage_delObjects("test-document")

        # Check that shared document is unindexed
        elements = self.portal.portal_catalog.searchResults(id="test-document")
        self.assertEqual(len(elements), 0)

    def test_shared_content_uuid(self):
        # Create shared document
        createContentInContainer(
            self.portal, "LIF", title="Assets", checkConstraints=False
        )
        createContentInContainer(self.portal.assets, "Document", title="Test document")

        root_uuid = IUUID(self.portal.assets["test-document"])
        shared_uuid = IUUID(self.portal.ca.recursos["test-document"])

        self.assertEqual(f"{root_uuid:s}-ca", shared_uuid)

    def test_moving_shared_content_to_lrf(self):
        # Create shared document
        createContentInContainer(
            self.portal, "LIF", title="Assets", checkConstraints=False
        )
        createContentInContainer(self.portal.assets, "Document", title="Test document")
        uuid = IUUID(self.portal.assets["test-document"])

        # Check that ghost is ghost
        self.assertTrue(
            is_language_independent(self.portal.ca.recursos["test-document"])
        )

        # Check is in the catalog
        brains = self.portal.portal_catalog.searchResults(UID=uuid)
        self.assertEqual(len(brains), 1)
        self.assertEqual(brains[0].getPath(), "/plone/assets/test-document")

        brains = self.portal.portal_catalog.searchResults(UID=f"{uuid:s}-ca")
        self.assertEqual(len(brains), 1)
        self.assertEqual(brains[0].getPath(), "/plone/ca/recursos/test-document")

        brains = self.portal.portal_catalog.searchResults(UID=f"{uuid:s}-es")
        self.assertEqual(len(brains), 1)
        self.assertEqual(brains[0].getPath(), "/plone/es/recursos/test-document")

        # MOVE!
        moved = multilingualMoveObject(self.portal.ca.recursos["test-document"], "ca")

        # Check that the old and the new uuid are the same
        moved_uuid = IUUID(self.portal.ca["test-document"])

        self.assertEqual(uuid, moved_uuid)
        self.assertFalse(is_language_independent(moved))

        # Check portal_catalog is updated after move
        brains = self.portal.portal_catalog.searchResults(UID=uuid)
        self.assertEqual(len(brains), 1)
        self.assertEqual(brains[0].getPath(), "/plone/ca/test-document")

        brains = self.portal.portal_catalog.searchResults(UID=f"{uuid:s}-ca")
        self.assertEqual(len(brains), 0)

        brains = self.portal.portal_catalog.searchResults(UID=f"{uuid:s}-es")
        self.assertEqual(len(brains), 0)

        # Check which translations it have
        self.assertEqual(ITranslationManager(moved).get_translations(), {"ca": moved})
        ITranslationManager(moved).add_translation("es")
        self.assertEqual(
            ITranslationManager(moved).get_translations(),
            {"ca": moved, "es": self.portal.es["test-document"]},
        )

        # Check that ghost is no longer ghost
        self.assertFalse(is_language_independent(self.portal.es["test-document"]))



from plone.app.multilingual import api
from plone.app.multilingual.dx.interfaces import ILanguageIndependentField
from plone.app.multilingual.interfaces import IPloneAppMultilingualInstalled
from plone.app.multilingual.testing import PAM_FUNCTIONAL_TESTING
from plone.app.relationfield.behavior import IRelatedItems
from plone.dexterity.fti import DexterityFTI
from plone.dexterity.utils import createContentInContainer
from plone.uuid.interfaces import IUUID
from Products.CMFCore.utils import getToolByName
from z3c.form.interfaces import IDataManager
from z3c.form.interfaces import IValidator
from z3c.relationfield import RelationValue
from zope.component import getMultiAdapter
from zope.event import notify
from zope.interface import alsoProvides
from zope.interface import noLongerProvides
from zope.lifecycleevent import modified
from zope.lifecycleevent import ObjectModifiedEvent
from zope.pagetemplate.interfaces import IPageTemplate
from zope.publisher.interfaces.browser import IDefaultBrowserLayer
from zope.schema._bootstrapinterfaces import RequiredMissing

import transaction
import unittest


class TestLanguageIndependentFieldOnAddTranslationForm(unittest.TestCase):
    layer = PAM_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        alsoProvides(self.request, IDefaultBrowserLayer)
        alsoProvides(self.request, IPloneAppMultilingualInstalled)

        fti = DexterityFTI("Feedback")
        fti.behaviors = (
            "plone.basic",
            "plone.namefromtitle",
            "plone.translatable",
        )
        fti.model_source = """\
<model xmlns="http://namespaces.plone.org/supermodel/schema"
       xmlns:lingua="http://namespaces.plone.org/supermodel/lingua">
  <schema>
    <field name="mandatory_feedback" type="zope.schema.Text"
           lingua:independent="true">
      <description />
      <required>True</required>
      <title>Mandatory feedback</title>
    </field>
  </schema>
</model>"""
        portal_types = getToolByName(self.portal, "portal_types")
        portal_types._setObject("Feedback", fti)

        self.document = createContentInContainer(
            self.portal["en"],
            "Feedback",
            checkConstraints=False,
            title="Test feedback",
            mandatory_feedback="This is a test",
        )

        # Call 'create_translation' to annotate request
        self.request.form["language"] = "ca"
        self.redirect = getMultiAdapter(
            (self.document, self.request), name="create_translation"
        )
        self.redirect()

        # Look up the ++addtranslation++ with annotated request in place
        self.view = self.portal["ca"].restrictedTraverse(
            "++addtranslation++" + IUUID(self.document)
        )
        self.view.update()
        self.field = self.view.form_instance.fields["mandatory_feedback"].field
        self.widget = self.view.form_instance.widgets["mandatory_feedback"]

    def test_field_is_required(self):
        self.assertTrue(self.field.required)

    def test_field_is_independent(self):
        self.assertTrue(ILanguageIndependentField.providedBy(self.field))

    def test_default_validator_raise_exception_on_independent_field(self):
        # Remove browser layer to get the default validator
        noLongerProvides(self.request, IPloneAppMultilingualInstalled)
        ###
        validator = getMultiAdapter(
            (
                self.view.context,
                self.view.request,
                self.view.form_instance,
                self.field,
                self.widget,
            ),
            IValidator,
        )
        self.assertNotEqual(
            str(validator.__class__.__name__), "LanguageIndependentFieldValidator"
        )
        self.assertRaises(RequiredMissing, validator.validate, None)

    def test_validator_pass_on_required_independent_field(self):
        validator = getMultiAdapter(
            (
                self.view.context,
                self.view.request,
                self.view.form_instance,
                self.field,
                self.widget,
            ),
            IValidator,
        )
        self.assertEqual(
            str(validator.__class__.__name__), "LanguageIndependentFieldValidator"
        )
        self.assertIsNone(validator.validate(None))

    def test_input_widget_renders_textarea(self):
        # Remove browser layer to get the default validator
        noLongerProvides(self.request, IPloneAppMultilingualInstalled)
        ###
        widget_template = getMultiAdapter(
            (
                self.view.context,
                self.view.request,
                self.view.form_instance,
                self.field,
                self.widget,
            ),
            IPageTemplate,
            name="input",
        )
        self.assertIn("<textarea", widget_template(self.widget))

    def test_input_widget_does_not_render_textarea_but_span(self):
        widget_template = getMultiAdapter(
            (
                self.view.context,
                self.view.request,
                self.view.form_instance,
                self.field,
                self.widget,
            ),
            IPageTemplate,
            name="input",
        )
        self.assertNotIn("<textarea", widget_template(self.widget))

    def test_field_copied_to_translation_when_modified(self):
        en_feedback = self.portal.en["test-feedback"]
        ca_feedback = api.translate(en_feedback, "ca")
        del en_feedback.REQUEST.translation_info

        en_feedback.mandatory_feedback = "modified"
        modified(en_feedback)
        self.assertEqual(ca_feedback.mandatory_feedback, "modified")

    def test_field_not_copied_if_not_modified(self):
        en_feedback = self.portal.en["test-feedback"]
        ca_feedback = api.translate(en_feedback, "ca")
        del en_feedback.REQUEST.translation_info

        # commit changes so far, so that we can check to see
        # if the translation was updated
        transaction.commit()
        modified(en_feedback)
        self.assertNotIn(ca_feedback, self.portal._p_jar._registered_objects)


class TestLanguageIndependentRelationField(unittest.TestCase):
    layer = PAM_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        alsoProvides(self.request, IDefaultBrowserLayer)
        alsoProvides(self.request, IPloneAppMultilingualInstalled)

        self.a_en = createContentInContainer(
            self.portal["en"], "Document", title="Test Document"
        )

        self.b_en = createContentInContainer(
            self.portal["en"], "Document", title="Another Document"
        )

        adapted = IRelatedItems(self.a_en)
        dm = getMultiAdapter((adapted, IRelatedItems["relatedItems"]), IDataManager)
        dm.set([self.b_en])

    def test_has_relation_list(self):
        adapted = IRelatedItems(self.a_en)

        bound = IRelatedItems["relatedItems"].bind(adapted)
        self.assertEqual(len(bound.get(adapted)), 1)

        value = bound.get(adapted)
        self.assertEqual(type(value[0]), RelationValue)

        dm = getMultiAdapter((adapted, IRelatedItems["relatedItems"]), IDataManager)
        self.assertEqual(dm.get(), [self.b_en])

    def test_relation_list_gets_copied(self):
        a_ca = api.translate(self.a_en, "ca")

        adapted = IRelatedItems(a_ca)

        bound = IRelatedItems["relatedItems"].bind(adapted)
        self.assertEqual(len(bound.get(adapted)), 1)

        value = bound.get(adapted)
        self.assertEqual(type(value[0]), RelationValue)

        dm = getMultiAdapter((adapted, IRelatedItems["relatedItems"]), IDataManager)
        self.assertEqual(dm.get(), [self.b_en])

    def test_relation_list_gets_translated(self):
        b_ca = api.translate(self.b_en, "ca")
        a_ca = api.translate(self.a_en, "ca")

        adapted = IRelatedItems(a_ca)

        bound = IRelatedItems["relatedItems"].bind(adapted)
        self.assertEqual(len(bound.get(adapted)), 1)

        value = bound.get(adapted)
        self.assertEqual(type(value[0]), RelationValue)

        dm = getMultiAdapter((adapted, IRelatedItems["relatedItems"]), IDataManager)
        self.assertEqual(dm.get(), [b_ca])

    def test_relation_list_gets_cleared(self):
        a_ca = api.translate(self.a_en, "ca")

        adapted = IRelatedItems(self.a_en)
        dm = getMultiAdapter((adapted, IRelatedItems["relatedItems"]), IDataManager)
        dm.set([])

        notify(ObjectModifiedEvent(self.a_en))

        adapted = IRelatedItems(a_ca)
        dm = getMultiAdapter((adapted, IRelatedItems["relatedItems"]), IDataManager)
        self.assertEqual(dm.get(), [])

    def test_copied_relation_list_gets_translated(self):
        a_ca = api.translate(self.a_en, "ca")
        b_ca = api.translate(self.b_en, "ca")

        # But only after self.a_en is modified (this is a feature, not a bug):
        notify(ObjectModifiedEvent(self.a_en))

        adapted = IRelatedItems(a_ca)

        bound = IRelatedItems["relatedItems"].bind(adapted)
        self.assertEqual(len(bound.get(adapted)), 1)

        value = bound.get(adapted)
        self.assertEqual(type(value[0]), RelationValue)

        dm = getMultiAdapter((adapted, IRelatedItems["relatedItems"]), IDataManager)
        self.assertEqual(dm.get(), [b_ca])



from plone.app.layout.viewlets.common import ViewletBase
from plone.app.multilingual.interfaces import ITranslatable
from plone.app.multilingual.interfaces import ITranslationManager
from plone.base.interfaces import ILanguage
from plone.memoize import ram
from Products.CMFCore.utils import getToolByName
from urllib.parse import quote_plus


def _cache_until_catalog_change(fun, self):
    catalog = getToolByName(self.context, "portal_catalog")
    key = "{0}{1}{2}"
    key = key.format(
        fun.__name__, catalog.getCounter(), "/".join(self.context.getPhysicalPath())
    )
    return key


class OneLanguageConfiguredNoticeViewlet(ViewletBase):
    """Notice the user that PAM is installed and only one language
    is configured.
    """

    available = False

    def render(self):
        if self.available:
            return self.index()

        return ""

    def update(self):
        lt = getToolByName(self.context, "portal_languages")
        supported = lt.getSupportedLanguages()
        self.available = len(supported) <= 1


class AddFormIsATranslationViewlet(ViewletBase):
    """Notice the user that this add form is a translation"""

    available = False

    def language(self):
        return self.lang

    def languages(self):
        """Returns list of languages."""
        self.tool = getToolByName(self.context, "portal_languages", None)
        if self.tool is None:
            return []

        languages = {
            lang: info
            for (lang, info) in self.tool.getAvailableLanguageInformation().items()
            if info["selected"]
        }

        return languages

    def language_name(self, lang_code):
        return self.languages().get(lang_code).get("native")

    def render(self):
        if self.available:
            return self.index()
        return ""

    def returnURL(self):
        # Get translation info for getting the translation source
        translation_info = getattr(self.request, "translation_info", {})
        translation_group = translation_info.get("tg")
        source_language = translation_info.get("source_language")

        if not (translation_group or source_language):
            return ""

        catalog = getToolByName(self.context, "portal_catalog")
        brains = catalog(Language=source_language, TranslationGroup=translation_group)
        if len(brains) != 1:
            return ""
        source = brains[0].getObject()

        # Get the factory
        types_tool = getToolByName(source, "portal_types")

        # Note: we don't check 'allowed' or 'available' here,
        # because these are slow. We assume the 'allowedTypes'
        # list has already performed the necessary calculations
        actions = types_tool.listActionInfos(
            object=self.context,
            check_permissions=False,
            check_condition=False,
            category="folder/add",
        )

        addActionsById = {a["id"]: a for a in actions}

        typeId = source.portal_type

        addAction = addActionsById.get(typeId, None)
        if addAction is None:
            url = None
        else:
            url = addAction["url"]
        if not url:
            url = "{}/createObject?type_name={}".format(
                source.absolute_url(),
                quote_plus(typeId),
            )
        return url

    def update(self):
        try:
            tg = self.request.translation_info["tg"]
        except AttributeError:
            return
        self.available = True
        if ITranslatable.providedBy(self.context):
            self.lang = ILanguage(self.context).get_language()
        else:
            self.lang = "NaN"
        catalog = getToolByName(self.context, "portal_catalog")
        query = {"TranslationGroup": tg}
        self.origin = catalog.searchResults(query)


class AddFormATIsATranslationViewlet(AddFormIsATranslationViewlet):
    # XXX move this class to archetypes multilingual!
    # btw., it is not used in here.
    """Notice the user that this AT add form is a translation"""

    def update(self):
        """It's only for AT on factory so we check"""
        factory = getToolByName(self.context, "portal_factory", None)
        if factory is None or not factory.isTemporary(self.context):
            return
        super(AddFormIsATranslationViewlet, self).update()


class AlternateLanguagesViewlet(ViewletBase):
    """Notice search engines about alternates languages of current
    content item
    """

    alternatives = []

    @ram.cache(_cache_until_catalog_change)
    def get_alternate_languages(self):
        """Cache relative urls only. If we have multilingual sites
        and multi domain site caching absolute urls will result in
        very inefficient caching. Build absolute url in template.
        """
        tm = ITranslationManager(self.context)
        catalog = getToolByName(self.context, "portal_catalog")
        results = catalog(TranslationGroup=tm.query_canonical())

        alternates = []
        for item in results:
            url = item.getURL()
            alternates.append(
                {
                    "lang": item.Language,
                    "url": url,
                }
            )

        return alternates

    def update(self):
        super().update()
        self.alternates = self.get_alternate_languages()

    @property
    def available(self):
        return len(self.alternates) > 1

    def render(self):
        if self.available:
            return self.index()
        return ""



from plone import api
from plone.app.multilingual.testing import (  # noqa
    PLONE_APP_MULTILINGUAL_INTEGRATION_TESTING,
)
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from Products.CMFPlone.utils import get_installer

import unittest


class TestUninstall(unittest.TestCase):
    layer = PLONE_APP_MULTILINGUAL_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        roles_before = api.user.get_roles(TEST_USER_ID)
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        self.installer = get_installer(self.portal)
        self.installer.uninstall_product("plone.app.multilingual")
        setRoles(self.portal, TEST_USER_ID, roles_before)

    def test_product_uninstalled(self):
        """Test if plone.app.multilingual is cleanly uninstalled."""
        self.assertFalse(self.installer.is_product_installed("plone.app.multilingual"))

    def test_browserlayer_removed(self):
        """Test that IPloneAppMultilingualInstalled is removed."""
        from plone.app.multilingual.interfaces import IPloneAppMultilingualInstalled
        from plone.browserlayer import utils

        self.assertNotIn(IPloneAppMultilingualInstalled, utils.registered_layers())

    def test_language_switcher_not_in_available_view_methods(self):
        self.assertNotIn(
            "language-switcher", self.portal.portal_types["Plone Site"].view_methods
        )

    def test_language_switcher_not_default_view_method(self):
        self.assertNotEqual(
            "language-switcher", self.portal.portal_types["Plone Site"].default_view
        )



from AccessControl import getSecurityManager
from AccessControl.SecurityManagement import newSecurityManager
from AccessControl.SecurityManagement import setSecurityManager
from AccessControl.users import UnrestrictedUser
from plone.app.multilingual.dx.interfaces import IDexterityTranslatable
from plone.app.multilingual.interfaces import ILanguageIndependentFieldsManager
from plone.app.multilingual.interfaces import IMultiLanguageExtraOptionsSchema
from plone.app.multilingual.interfaces import IPloneAppMultilingualInstalled
from plone.app.multilingual.interfaces import ITranslationManager
from plone.base.interfaces import ILanguage
from plone.dexterity.interfaces import IDexterityFTI
from plone.registry.interfaces import IRegistry
from Products.CMFCore.utils import getToolByName
from zope.component import getUtility
from zope.component import queryAdapter
from zope.event import notify
from zope.globalrequest import getRequest
from zope.lifecycleevent import Attributes
from zope.lifecycleevent import ObjectModifiedEvent
from zope.lifecycleevent.interfaces import IObjectModifiedEvent


class LanguageIndependentModifier:
    """Class to handle dexterity editions."""

    def __call__(self, content, event):
        """Called by the event system."""
        request = getattr(event.object, "REQUEST", getRequest())

        if not IPloneAppMultilingualInstalled.providedBy(request):
            return

        translation_info = getattr(request, "translation_info", {})

        if "tg" in translation_info.keys():
            # In case it's a on the fly translation avoid
            return

        if IDexterityTranslatable.providedBy(content):
            self.canonical = ITranslationManager(content).query_canonical()

            if (
                event.descriptions
                and len(event.descriptions) > 1
                and event.descriptions[1] == self.canonical
            ):
                return

            if IObjectModifiedEvent.providedBy(event):
                self.handle_modified(content)

    def bypass_security_checks(self):
        registry = getUtility(IRegistry)

        # BBB for lrf-branch
        field = registry.records.get(
            IMultiLanguageExtraOptionsSchema.__identifier__
            + ".bypass_languageindependent_field_permission_check"
        )

        return field and field.value or False

    def handle_modified(self, content):
        fieldmanager = ILanguageIndependentFieldsManager(content)
        if not fieldmanager.has_independent_fields():
            return

        sm = getSecurityManager()
        try:
            # Do we have permission to sync language independent fields?
            if self.bypass_security_checks():
                # Clone the current user and assign a new editor role to
                # allow edition of all translated objects even if the
                # current user whould not have permission to do that.
                tmp_user = UnrestrictedUser(
                    sm.getUser().getId(),
                    "",
                    [
                        "Editor",
                    ],
                    "",
                )

                # Wrap the user in the acquisition context of the portal
                # and finally switch the user to our new editor
                acl_users = getToolByName(content, "acl_users")
                tmp_user = tmp_user.__of__(acl_users)
                newSecurityManager(None, tmp_user)

            # Copy over all language independent fields
            transmanager = ITranslationManager(content)
            for translation in self.get_all_translations(content):
                trans_obj = transmanager.get_translation(translation)
                if fieldmanager.copy_fields(trans_obj):
                    self.reindex_translation(trans_obj)
        finally:
            # Restore the old security manager
            setSecurityManager(sm)

    def reindex_translation(self, translation):
        """Once the modification is done, reindex translation"""
        translation.reindexObject()

        fti = getUtility(IDexterityFTI, name=translation.portal_type)
        schema = fti.lookupSchema()
        descriptions = Attributes(schema)
        # where is this information needed?
        # XXX behaviors need to be considered here
        # use plone.dexterity.utils.iterSchemata or similiar

        # Pass the canonical object as a event description
        notify(ObjectModifiedEvent(translation, descriptions, self.canonical))

    def get_all_translations(self, content):
        """Return all translations excluding the just modified content"""
        content_lang = queryAdapter(content, ILanguage).get_language()
        translations = ITranslationManager(content).get_translated_languages()
        while content_lang in translations:
            translations.remove(content_lang)
        return translations


handler = LanguageIndependentModifier()



from plone.app.multilingual.dx.directives import languageindependent
from plone.app.multilingual.dx.interfaces import ILanguageIndependentField
from plone.app.multilingual.testing import PAM_INTEGRATION_TESTING
from plone.supermodel import model
from plone.supermodel.utils import mergedTaggedValueList
from zope import schema
from zope.interface import Interface

import unittest


class TestDirectives(unittest.TestCase):
    layer = PAM_INTEGRATION_TESTING

    def test_schema_directives_store_tagged_values(self):
        """Test, if the schema directive values are stored as tagged values."""

        class IDummy(model.Schema):
            """Dummy schema class."""

            languageindependent("foo")
            foo = schema.TextLine(title="Foo")

        self.assertEqual(
            [(Interface, "foo", "true")],
            mergedTaggedValueList(IDummy, languageindependent.key),
        )

        self.assertTrue(ILanguageIndependentField.providedBy(IDummy["foo"]))

    def test_inherited_schema_still_has_tagged_value(self):
        """An inherited schema should still have the tagged value information
        inherited from its superclass.
        """

        class IFoo(model.Schema):
            """Class with a searchable field"""

            languageindependent("baz")
            baz = schema.TextLine(title="baz")

        class IBar(IFoo):
            """Schema class which inherits a field from IFoo."""

        self.assertEqual(
            [(Interface, "baz", "true")],
            mergedTaggedValueList(IFoo, languageindependent.key),
        )
        self.assertTrue(ILanguageIndependentField.providedBy(IFoo["baz"]))

        self.assertEqual(
            [(Interface, "baz", "true")],
            mergedTaggedValueList(IBar, languageindependent.key),
        )
        self.assertTrue(ILanguageIndependentField.providedBy(IBar["baz"]))



from AccessControl.SecurityManagement import getSecurityManager
from Acquisition import aq_chain
from borg.localrole.interfaces import IFactoryTempFolder
from plone.app.layout.navigation.interfaces import INavigationRoot
from plone.app.multilingual.browser.selector import addQuery
from plone.app.multilingual.browser.selector import NOT_TRANSLATED_YET_TEMPLATE
from plone.app.multilingual.interfaces import ILanguageRootFolder
from plone.app.multilingual.interfaces import IMultiLanguageExtraOptionsSchema
from plone.app.multilingual.interfaces import ITG
from plone.app.multilingual.interfaces import ITranslatable
from plone.app.multilingual.interfaces import ITranslationManager
from plone.app.multilingual.manager import TranslationManager
from plone.i18n.interfaces import INegotiateLanguage
from plone.i18n.locales.interfaces import IContentLanguageAvailability
from plone.registry.interfaces import IRegistry
from Products.CMFCore.interfaces import ISiteRoot
from Products.CMFCore.utils import getToolByName
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.component.hooks import getSite
from zope.interface import implementer
from zope.publisher.interfaces import IPublishTraverse
from zope.publisher.interfaces import NotFound


@implementer(IPublishTraverse)
class universal_link(BrowserView):
    """Redirects the user to the negotiated translated page
    based on the user preferences in the user's browser.
    """

    def __init__(self, context, request):
        super().__init__(context, request)
        self.tg = None
        self.lang = None

    def publishTraverse(self, request, name):
        if self.tg is None:  # ../@@universal-link/translationgroup
            self.tg = name
        elif self.lang is None:  # ../@@universal-link/translationgroup/lang
            self.lang = name
        else:
            raise NotFound(self, name, request)

        return self

    def getDestination(self):
        # Look for the element
        url = None
        # We check if it's shared content
        query = {"TranslationGroup": self.tg}
        ptool = getToolByName(self.context, "portal_catalog")
        brains = ptool.searchResults(query)
        is_shared = False
        for brain in brains:
            if "-" in brain.UID:
                is_shared = True
                brain_to_use = brain
                break
        if is_shared:
            target_uid = brain_to_use.UID.split("-")[0]
            if self.lang:
                target_uid += "-" + self.lang
            else:
                # The negotiated language
                language = getMultiAdapter(
                    (self.context, self.request), INegotiateLanguage
                ).language
                target_uid += "-" + language
            results = ptool.searchResults(UID=target_uid)
            if len(results) > 0:
                url = results[0].getURL()
            return url
        # Normal content
        query = {"TranslationGroup": self.tg}
        if self.lang:
            query = {"TranslationGroup": self.tg, "Language": self.lang}
        else:
            # The negotiated language
            language = getMultiAdapter(
                (self.context, self.request), INegotiateLanguage
            ).language
            query = {"TranslationGroup": self.tg, "Language": language}

        # Comparison to plone/app/multilingual/browser/setup.py#L129
        if query.get("Language") == "id-id":
            query["Language"] = "id"

        results = ptool.searchResults(query)
        if len(results) > 0:
            url = results[0].getURL()
        return url

    def __call__(self):
        url = self.getDestination()
        if not url:
            root = getToolByName(self.context, "portal_url")
            url = root()
        self.request.RESPONSE.redirect(url)


class selector_view(universal_link):
    def getDialogDestination(self):
        """Get the "not translated yet" dialog URL.
        It's located on the root and shows the translated objects of that TG
        """
        dialog_view = NOT_TRANSLATED_YET_TEMPLATE
        postpath = False
        # The dialog view shouldn't appear on the site root
        # because that is untraslatable by default.
        # And since we are mapping the root on itself,
        # we also do postpath insertion (@@search case)

        root = getToolByName(self.context, "portal_url")
        url = root() + dialog_view + "/" + self.tg
        return self.wrapDestination(url, postpath=postpath)

    def getParentChain(self, context):
        # XXX: switch it over to parent pointers if needed
        return aq_chain(context)

    def getClosestDestination(self):
        """Get the "closest translated object" URL."""
        # We should travel the parent chain using the catalog here,
        # but I think using the acquisition chain is faster
        # (or well, __parent__ pointers) because the catalog
        # would require a lot of queries, while technically,
        # having done traversal up to this point you should
        # have the objects in memory already

        # As we don't have any content object we are going to look
        # for the best option

        site = getSite()
        root = getToolByName(site, "portal_url")
        ltool = getToolByName(site, "portal_languages")

        # We are using TranslationManager to get the translations of a
        # string tg
        try:
            manager = TranslationManager(self.tg)
            languages = manager.get_translations()
        except AttributeError:
            languages = []
        if len(languages) == 0:
            # If there is no results there are no translations
            # we move to portal root
            return self.wrapDestination(root(), postpath=False)

        # We are going to see if there is the preferred language translation
        # Otherwise we get the first as context to look for translation
        prefered = ltool.getPreferredLanguage(self.request)
        if prefered in languages:
            context = languages[prefered]
        else:
            context = languages[list(languages.keys())[0]]

        checkPermission = getSecurityManager().checkPermission
        chain = self.getParentChain(context)
        for item in chain:
            if ISiteRoot.providedBy(item) and not ILanguageRootFolder.providedBy(item):
                # We do not care to get a permission error
                # if the whole of the portal cannot be viewed.
                # Having a permission issue on the root is fine;
                # not so much for everything else so that is checked there
                return self.wrapDestination(item.absolute_url())
            try:
                canonical = ITranslationManager(item)
            except TypeError:
                if not ITranslatable.providedBy(item):
                    # In case there it's not translatable go to parent
                    # This solves the problem when a parent is not
                    # ITranslatable
                    continue
                else:
                    raise
            translation = canonical.get_translation(self.lang)
            if translation and (
                INavigationRoot.providedBy(translation)
                or bool(checkPermission("View", translation))
            ):
                # Not a direct translation, therefore no postpath
                # (the view might not exist on a different context)
                return self.wrapDestination(translation.absolute_url(), postpath=False)
        # Site root's the fallback
        return self.wrapDestination(root(), postpath=False)

    def wrapDestination(self, url, postpath=True):
        """Fix the translation url appending the query
        and the eventual append path.
        """
        if postpath:
            url += self.request.form.get("post_path", "")
        return addQuery(self.request, url, exclude=("post_path"))

    def __call__(self):
        url = self.getDestination()
        if url:
            # We have a direct translation, full wrapping
            url = self.wrapDestination(url)
        else:
            registry = getUtility(IRegistry)
            policies = registry.forInterface(
                IMultiLanguageExtraOptionsSchema, prefix="plone"
            )
            if policies.selector_lookup_translations_policy == "closest":
                url = self.getClosestDestination()
            else:
                url = self.getDialogDestination()
            # No wrapping cause that's up to the policies
            # (they should already have done that)
        self.request.response.redirect(url)


@implementer(IPublishTraverse)
class not_translated_yet(BrowserView):
    """View to inform user that the view requested is not translated yet,
    and shows the already translated related content.
    """

    __call__ = ViewPageTemplateFile("templates/not_translated_yet.pt")

    def __init__(self, context, request):
        super().__init__(context, request)
        self.tg = None

    def publishTraverse(self, request, name):
        if self.tg is None:  # ../@@not_translated_yet/translationgroup
            self.tg = TranslationManager(name)
        else:
            raise NotFound(self, name, request)
        return self

    def already_translated(self):
        if self.tg is not None:
            # Need to check if the user has permission
            return self.tg.get_restricted_translations().items()
        else:
            return []

    def has_any_translation(self):
        if self.tg is not None:
            return len(self.tg.get_restricted_translations().items()) > 0
        else:
            return 0

    def language_name(self, lang=None):
        """Get the current language native name"""
        if lang is None:
            lang_code = self.request.get("set_language")
        else:
            lang_code = lang
        util = getUtility(IContentLanguageAvailability)
        data = util.getLanguages(True)
        lang_info = data.get(lang_code)
        if lang_info is None:
            return None
        return lang_info.get("native", None) or lang_info.get("name")


class TGView(BrowserView):
    """A simple browser view that renders the TG of its context"""

    def __call__(self):
        return str(ITG(self.context, ""))



from BTrees.OOBTree import OOBTree
from plone.dexterity.content import Container
from Products.CMFCore.utils import getToolByName
from zope.component.hooks import getSite


def connect_with_others(ob, event):
    pc = getToolByName(getSite(), "portal_catalog")
    results = pc.unrestrictedSearchResults(portal_type="LIF")
    for result in results:
        lif = result._unrestrictedGetObject()
        ob._tree = lif._tree
        ob._count = lif._count
        ob._mt_index = lif._mt_index
        if not hasattr(lif, "__annotations__"):
            lif.__annotations__ = OOBTree()
        ob.__annotations__ = lif.__annotations__
        break



from Acquisition import aq_base
from plone.app.multilingual.browser.setup import SetupMultilingualSite
from plone.app.multilingual.browser.vocabularies import (  # noqa: E501
    AllContentLanguageVocabulary,
)
from plone.app.multilingual.interfaces import ATTRIBUTE_NAME
from plone.app.multilingual.interfaces import IPloneAppMultilingualInstalled
from plone.app.multilingual.interfaces import ITG
from plone.app.multilingual.interfaces import ITranslatable
from plone.app.multilingual.testing import PAM_INTEGRATION_PRESET_TESTING
from plone.app.multilingual.testing import PAM_INTEGRATION_TESTING
from Products.CMFCore.utils import getToolByName
from zope.interface import alsoProvides

import unittest


class TestSetupMultilingualSite(unittest.TestCase):
    """Testing multilingual site without predefined languages."""

    layer = PAM_INTEGRATION_TESTING

    def setUp(self):
        """Setting up the test."""
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        self.language_tool = getToolByName(self.portal, "portal_languages")
        self.languages = self.language_tool.getSupportedLanguages()
        alsoProvides(self.layer["request"], IPloneAppMultilingualInstalled)

    def test_single_language(self):
        """Only one language is set."""
        self.assertEqual(len(self.languages), 1)

    def test_portal_has_tg_attribute(self):
        """The site root should have the TG attribute set after installing"""
        tg_attribute = getattr(aq_base(self.portal), ATTRIBUTE_NAME, None)
        self.assertIsNotNone(tg_attribute)

    def test_no_languagefolder_created(self):
        """On a single language no folder creation is done."""
        portal_objects = self.portal.objectIds()
        for lang in self.languages:
            self.assertNotIn(lang, portal_objects)

    def test_all_supported_languages(self):
        """There was a language which code is 'id'.
        This broke the root language folder setup process.
        To get rid of that the folder is 'id-id'.
        """
        all_langs = AllContentLanguageVocabulary()(self.portal)
        for lang in all_langs:
            self.language_tool.addSupportedLanguage(lang.value)

        workflow_tool = getToolByName(self.portal, "portal_workflow")
        workflow_tool.setDefaultChain("simple_publication_workflow")

        setup_tool = SetupMultilingualSite()
        setup_tool.setupSite(self.portal)

        portal_objects = self.portal.objectIds()

        for lang in all_langs.by_value.keys():
            if lang == "id":
                self.assertIn("id-id", portal_objects)
            else:
                self.assertIn(lang, portal_objects)

    def test_type_of_language_folders(self):
        """The created objects have to be 'Language Root Folder'."""
        all_langs = AllContentLanguageVocabulary()(self.portal)
        for lang in all_langs:
            self.language_tool.addSupportedLanguage(lang.value)

        workflow_tool = getToolByName(self.portal, "portal_workflow")
        workflow_tool.setDefaultChain("simple_publication_workflow")

        setup_tool = SetupMultilingualSite()
        setup_tool.setupSite(self.portal)

        for lang in all_langs.by_value.keys():
            if lang == "id":
                self.assertEqual(self.portal.get("id-id").portal_type, "LRF")
            else:
                self.assertEqual(self.portal.get(lang).portal_type, "LRF")


class TestSetupMultilingualPresetSite(unittest.TestCase):
    """Testing multilingual site with predefined languages."""

    layer = PAM_INTEGRATION_PRESET_TESTING

    def setUp(self):
        """Setting up the test."""
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        self.language_tool = getToolByName(self.portal, "portal_languages")
        self.languages = self.language_tool.getSupportedLanguages()
        alsoProvides(self.layer["request"], IPloneAppMultilingualInstalled)

    def test_language_folders_created(self):
        """Available languages are: 'en', 'ca', 'es'.
        After setup has run there should be
        new objects in the portal named after the languages.
        """
        portal_objects = self.portal.objectIds()
        for lang in self.languages:
            self.assertIn(lang, portal_objects)

    def test_type_of_language_folders(self):
        """The created objects have to be 'Language Root Folder'."""
        for lang in self.languages:
            self.assertEqual(self.portal.get(lang).portal_type, "LRF")



"""Setup plone.app.multilingual."""
from setuptools import find_packages
from setuptools import setup


setup(
    # metadata in setup.cfg
    packages=find_packages("src"),
    package_dir={"": "src"},
    namespace_packages=["plone", "plone.app"],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        "plone.i18n>=4.0.4",
        "setuptools",
    ],
    extras_require={
        "archetypes": [
            "archetypes.multilingual",
        ],
        "test": [
            "plone.app.testing[robot]",
            "plone.app.robotframework",
        ],
    },
    entry_points="""
    # -*- Entry points: -*-
    [z3c.autoinclude.plugin]
    target = plone
    """,
)
